"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

from PyQt5 import QtWidgets

from .rcallback import RCallback
from pyjamas.rimage.rimclassifier.rimclassifier import rimclassifier


class RCBSaveClassifier(RCallback):
    def cbSaveClassifier(self, filename: str = None, theclassifier: rimclassifier = None) -> bool:  # Handle IO errors.
        if theclassifier is None or theclassifier is False:
            if self.pjs.batch_classifier.image_classifier is None:
                return False
            else:
                theclassifier = self.pjs.batch_classifier.image_classifier

        # Get file name.
        if filename is None or filename is False: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            defaultname, _ = os.path.splitext(os.path.basename(self.pjs.filename))
            fname = QtWidgets.QFileDialog.getSaveFileName(None, 'Save classifier ...', self.pjs.cwd + os.sep + defaultname,
                                                          filter='PyJAMAS classifier (*' + rimclassifier.CLASSIFIER_EXTENSION + ')')  # fname[0] is the full filename, fname[1] is the filter used.

            # If cancel ...
            if fname == '':
                return False

            # else ...
            else:
                filename = fname[0]

        if filename[-4:] != rimclassifier.CLASSIFIER_EXTENSION:
            filename = filename + rimclassifier.CLASSIFIER_EXTENSION

        self.pjs.cwd = os.path.dirname(filename)

        theclassifier.save_classifier(filename)

        self.pjs.statusbar.showMessage(f'Saved {filename}.')

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

from PyQt5 import QtWidgets

from .rcallback import RCallback


class RCBSetCWD(RCallback):
    def cbSetCWD(self, folder_name: str = '') -> bool:
        if folder_name == '' or folder_name is False: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            folder_name = QtWidgets.QFileDialog.getExistingDirectory(None, 'Save files to folder ...', self.pjs.cwd)  # fname[0] is the full filename, fname[1] is the filter used.

            # If cancel ...
            if folder_name == '':
                return False

        if os.path.exists(folder_name):
            self.pjs.cwd = os.path.abspath(folder_name)
            self.pjs.statusbar.showMessage(f"Working folder set to {self.pjs.cwd}.")

            return True

        else:
            return False


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
from typing import List

import numpy
from PyQt5 import QtCore, QtWidgets

from pyjamas.pjsthreads import Thread
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimutils import rimutils
from pyjamas.rutils import RUtils


class RCBImage(RCallback):
    CW: int = 90
    CCW: int = -90
    UP_DOWN: int = 1
    LEFT_RIGHT: int = 2

    def cbRotateImage(self, direction: int = CW) -> bool:
        if direction == self.CW:
            # self.pjs.slices = numpy.asarray([numpy.rot90(x, -1) for x in self.pjs.slices]) # An order of magnitude slower than below.
            self.pjs.slices = numpy.rot90(self.pjs.slices, -1, (1, 2))

        elif direction == self.CCW:
            # self.pjs.slices = numpy.asarray([numpy.rot90(x) for x in self.pjs.slices]) # An order of magnitude slower than below.
            self.pjs.slices = numpy.rot90(self.pjs.slices, 1, (1, 2))

        self.pjs.initImage()

        return True

    def cbFlipImage(self, direction: int = LEFT_RIGHT) -> bool:
        if direction == self.LEFT_RIGHT:
            # self.pjs.slices = numpy.flip(self.pjs.slices, 2)  # Order or magnitude slower than the code below.
            self.pjs.slices = self.pjs.slices[..., ::-1]
        elif direction == self.UP_DOWN:
            # self.pjs.slices = numpy.fliplr(self.pjs.slices)  # Could have used flip with parameter 1, but this is faster. Unfortunately, there is no fast function to flip with parameter 2. Still, slower than below.
            self.pjs.slices = self.pjs.slices[..., ::-1, :]

        self.pjs.initImage()

        return True

    def cbMIPImage(self, slice_list: List[int]) -> bool:
        if slice_list is False:
            slice_list_str, ok_pressed = QtWidgets.QInputDialog.getText(None, "Maximum intenstiy projection", "Enter a range of slices (e.g. 0, 1, 4-8, 15): ", QtWidgets.QLineEdit.Normal, "")

            if not ok_pressed:
                return False

            if slice_list_str == '':
                slice_list = list(range(self.pjs.n_frames))
            else:
                slice_list = RUtils.parse_range_list(slice_list_str)

        # This line here is necessary: for some mysterious reason, if doing an MIP from a slice
        # other than the first one, there is an error that I was unable to debug, but seemed related
        # to the Qt backend based on the debugger error.
        """/Users/rodrigo/src/pyjamas/pyjamas/rimage/rimutils.py:641: RuntimeWarning: divide by zero encountered in true_divide
            sc = (maximum - minimum) / (high - low)
            /Users/rodrigo/src/pyjamas/pyjamas/rimage/rimutils.py:648: RuntimeWarning: invalid value encountered in multiply
            image_out = image_out * sc"""
        self.pjs._cbTimeSlider.cbGoTo(0)

        self.pjs.slices = rimutils.mip(self.pjs.slices[slice_list])

        self.pjs.initImage()

        return True

    def cbInvertImage(self) -> bool:
        self.pjs.slices = rimutils.invert(self.pjs.slices)
        self.pjs.prepare_image()

        self.pjs.displayData()

        return True

    def cbRegisterImage(self) -> bool:
        working_slice = self.pjs.curslice
        self.pjs.slices, distances = rimutils.register(self.pjs.slices, RUtils.pjsfiducials_to_array(self.pjs.fiducials), self.pjs.curslice)
        self.shift_annotations(distances)
        self.pjs.prepare_image()
        self.pjs._cbTimeSlider.cbGoTo(working_slice)

        self.pjs.displayData()

        return True

    # Inspiration for multithreading from: # Shamelessly stolen from https://www.mfitzp.com/article/multithreading-pyqt-applications-with-qthreadpool/
    def cbPlay(self) -> bool: # oh_no
        athread = Thread(self.play_movie)
        athread.kwargs['progress_callback'] = athread.signals.progress
        #athread.signals.result.connect(self.print_output)
        athread.signals.finished.connect(self.thread_complete)
        athread.signals.progress.connect(self.progress_fn)

        self.pjs.threadpool.start(athread)

        return True

    def play_movie(self, progress_callback) -> bool: #execute_this_fn
        start = time.time()
        period = 1.0 / self.pjs.fps

        while self.pjs.curslice < self.pjs.n_frames - 1:
            if (time.time() - start) > period:
                start += period
                # Because this is the function that will run in a thread, it cannot
                # manipulate the gui (or errors will happen - gui manipulations must
                # be in the thread that owns the gui). Instead, we can emit a signal
                # when it corresponds, and the signal can be slotted in a function
                # in the main thread that modifies the gui.
                #self.pjs._cbNextFrame.cbNextFrame()
                progress_callback.emit(self.pjs.curslice)

        return True

    def progress_fn(self, n):
        self.pjs._cbNextFrame.cbNextFrame()

    def thread_complete(self):
        self.pjs._cbTimeSlider.cbGoTo(0)

    def shift_annotations(self, translation_vector: numpy.ndarray) -> bool:
        if translation_vector is False or translation_vector is None:
            return False

        for slice_index in range(self.pjs.n_frames):
            theshift = translation_vector[slice_index].astype(int)

            for fiducial_index, _ in enumerate(self.pjs.fiducials[slice_index]):
                 self.pjs.fiducials[slice_index][fiducial_index] += theshift

            for a_polyline in self.pjs.polylines[slice_index]:
                a_polyline.translate(QtCore.QPointF(theshift[0], theshift[1]))

        return True


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gzip
import os
import pickle

import numpy
from PyQt5 import QtCore, QtGui, QtWidgets

import pyjamas.pjscore as pjscore
from pyjamas.rcallbacks.rcallback import RCallback


class RCBLoadAnnotations(RCallback):
    def cbLoadAnnotations(self, filename: str = None, image_file: str = None):  # Handle IO errors
        """

        :param filename:
        :param image_file: path to an image to be loaded with the annotation file. None if no image is to be loaded.
        '' to create a black image.
        :return:
        """

        # Get file name.
        if filename is None or filename is False or filename == '': # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getOpenFileName(None, 'Load classifier ...', self.pjs.cwd,
                                                          filter='PyJAMAS data (*' + pjscore.PyJAMAS.data_extension + ')')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        # Open file name and read annotations.
        fh = None

        try:
            fh = gzip.open(filename, "rb")
            fiducials = pickle.load(fh)
            polyline_list = pickle.load(fh)

        except (IOError, OSError) as ex:
            if fh is not None:
                fh.close()

            print(ex)

        if image_file == '':  # Create a blank image.
            # Find maxx, maxy, and maxz, and create a numpy.ndarray with those dimensions + 1.
            maxz = len(fiducials)
            maxx = 0
            maxy = 0

            # Fiducials and polylines are stored as (x, y) coordinates.
            # numpy.ndarrays are created with (slices, rows, cols).
            for slice_fiducials in fiducials:
                for afiducial in slice_fiducials:
                    maxx = max(maxx, afiducial[0])
                    maxy = max(maxy, afiducial[1])

            for slice_polylines in polyline_list:
                for apolyline in slice_polylines:
                    for afiducial in apolyline:
                        maxx = max(maxx, afiducial[0])
                        maxy = max(maxy, afiducial[1])

            virtual_image = numpy.zeros((maxz, int(maxy+1), int(maxx+1)), dtype=int)
            self.pjs._cbLoadTS.cbLoadArray(virtual_image)

        elif image_file is not False and image_file is not None:
            self.pjs._cbLoadTS.cbLoadTimeSeries(image_file)

        self.pjs.fiducials = fiducials

        self.pjs.polylines = [[] for i in range(self.pjs.n_frames)]
        for i, theframepolylines in enumerate(polyline_list):
            for j, thepolyline in enumerate(theframepolylines):
                if thepolyline != [[]]:
                    self.pjs.polylines[i].append(QtGui.QPolygonF())
                    for thepoint in thepolyline:
                        self.pjs.polylines[i][-1].append(QtCore.QPointF(thepoint[0], thepoint[1]))

        self.pjs.repaint()

        # Modify current path.
        self.pjs.cwd = filename[0:filename.rfind(os.sep)]

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from functools import partial
import os, sys
from typing import Tuple

import numpy
from PyQt5 import QtCore, QtGui, QtWidgets
from skimage.measure import points_in_poly

from pyjamas.dragdropmainwindow import DragDropMainWindow
from pyjamas.rimage.rimclassifier.batchclassifier import BatchClassifier
import pyjamas.rimage.rimcore as rimcore
from pyjamas.rimage.rimutils import rimutils
from pyjamas.rutils import RUtils


class PyJAMAS(QtCore.QObject):
    '''
    PyJAMAS is Just A More Awesome Siesta.

    Uses semantic versioning: https://semver.org/.
    Currently on 0.1.0.

    PyJAMAS() creates a PyJAMAS user interface.

    PROPERTIES -----------------------
    (Created with the following code:
    thekeys = list(self.__dict__.keys())
    thetypes = [str(type(x)) for x in self.__dict__.values()]
    [(x, thetypes[ind]) for ind, x in enumerate(thekeys)]
    )


     todo:
     - Follow track fiducials example to run in a thread with segmentation and classification options.
     - Should crop into new window launch a new thread?
     - Batch measure movies (polygon_analysis_script_wh).
     - Random forests, Neural networks, K-means clustering (https://www.kdnuggets.com/2019/05/guide-k-means-clustering-algorithm.html)
     - Move polylines with cursor.
     - Open and closed trajectories.
     - Create one class per callback function, and use a common name ('cb'?) for the callback function in each case?
     - Standardize features for classifiers (see Chapter 2 in Raschka). Uses sklearn's StandardScaler.
     - PIV-based image registration.
     - Add PIV to track_fiducials to improve the results (as an option).
     - Improve watershed: are we sure that the images that we feed into the watershed functions are identical in SIESTA and PyJAMAS? For instance, do the Gaussian smoothing functions result in indentical images when identical parameters are used?
     - Add option to quantify trajectory profile.
     - Merge rcbnextframe, rcbprevframe and rcbtimeslider.
     - Scrollable status bar: https://stackoverflow.com/questions/28655198/best-way-to-display-logs-in-pyqt
     - Adjust contrast (currently linear stretch between 0 and max: see pjscore.displayData, add options for custom min/max values, based on displaying the histogram and the resulting image within the dialog).
     - Save display (converts to 8b).
     - Export movie.
     - Add ROI options to find_seeds, expand_seeds, propagate_seeds, etc.
     - cbFindSeeds allows selection of thresholding method, and displays all possible outcomes for one image using skimage.filters import try_all_threshold
     - Menus disappear when opening a matplotlib figure - can be solved my minimizing and remaximizing the main window.
     - Remove shapely dependency by using skimage.measure.points_in_poly??? Problem are the exterior, centroid and area functions of the shapely.polygon class
     - Undo button: store n copies of pyjamas in a fifo cue. n is determined by an option. __copy__ and __deepcopy__ methods are already "drafted".
     - Parallelize non_max_suppression in rimclassifier.
     - Many of the options in the Image dialog get rid of all annotations (e.g. invert, which probably should not). At least issue a warning the first time.
     - Instead of using timepoints dialog, use rutils.parse_range_list.
     - Move dragdropmainwindow to dialogs module (?).
     - Bug: keyboard (mouse?) does not work on original pyjamas after crop into new window.
     - Bug: some times scrollbar not properly repositioned after zoom/window resize (when image is big or small?).
       BUT IT WORKS BEAUTIFULLY IF YOU ADD A FIDUCIAL IN THE IMAGE BEFORE RESIZING.
     - Add image/path parameter to the constructor of PyJAMAS (to avoid loading logo). Use in rcbcrop for crop into new pyjamas.
     - Conway's game of life for initial image (start with Pyjamas pattern). Add a "divertimento" that binarizes the image and
     uses it as the initial configuration for Conway's. Need "binarize" option.
     - Move everything in rimcore to rimutils (as class methods). Get rid of rimcore. Alternatively, implement
     rimutils methods as instance methods in rimcore.
     - centerSeeds is SLOW!!! Can it be optimized? Check Cython's pure python mode - https://cython.readthedocs.io/en/latest/src/tutorial/pure.html
     - MEDUSA: http://scikit-image.org/docs/dev/auto_examples/edges/plot_active_contours.html
     - Distance-transform based watershed.
     - Filter output section in rimutils.flow.
     - Add option for floating point fps (e.g. 0.5). Currently only integers are acceptable.
     - Add options to do MIP and concatenation of files in folders, so that you can get rid of ImageJ.
     - Add option to change brightness and contrast (a la Fiji).
     - Consider splitting rcallbacks module into submodules (one per menu?).
     - Add option to add a polygon with the alpha shape of the fiducials (ask for the value of alpha)
     - See scipy.spatial.ConvexHull and modify RUtils.concave_hull to return the points in order.
     - In expand seeds, compare images pixel by pixel with Matlab. Are the results in PyJAMAS as good as SIESTA's?
     - Consider storing all callback classes into a single class, such that one can use a = PyJAMAS(), a.callbacks.load_time_series(filename)
     (rather than the current a.load_time_series_object.load_time_series(filename)).
     - Consider merging rcbimage (flip, rotate, etc.) and rcbsaveimage.
     - Circular import: right now PyJAMAS imports rcallbacks inside a method because if imported at the beginning,
     rcallbacks also imports PyJAMAS at the beginning and that leads to an issue: the from rcallbacks import * at the beginning
     of pyjamas jumps to rcallbacks.py, and the import pyjamas at the beginning of rcallbacks jumps back to pyjamas. When
     the execution realizes the circular loop, it continues a reaches calls to rcallbacks that have not been defined because ...
     Is this solved in Python 3.7 con future import? Try to search for importing an undefined package and circular imports in stackoverflow.
     - (Same as above?) Cyclic import between pyjamas and rcallback ends up up a module 'pyjamas' has no attribute 'PyJAMAS'. So I
       delayed the import of rcallbacks until the setupUI method in this class.
     - Fix the problem with skimage io.imread loading 4-page images in rcbloadtimeseries (documented there).
     - Single fiducial seed propagation crashes.
     - add gradient option to flow calculations.
     - Write rimcore.makeGraph in cython (how to debug?).
     - Sparse matrices in cython (see https://www.evernote.com/shard/s448/nl/82435167/75e1d48a-fde9-4f5b-8b4c-7eebd3a2e7c9/)
     - Deep learning-based image segmentation
     - Raise exceptions instead of returning nones
     - Profile with quantifiedcode.
     - Add a "BATCH" menu to process multiple images
     - Quantify laser cutting (local_all_these_movies_time).
     - Quantify and plot polygon data (polygon_analysis_script_wh?).
     - LiveTrajectories.
     - LiveA* (calling mex code?).
     - QuPID
     - Zoom out -> when you load an image, zoom it out if it is too big to fit in the screen.
     - Docstring document of properties, methods and classes. Use sphinx: http://www.sphinx-doc.org/en/stable/tutorial.html
     - Copy & paste polylines.
     - Is there any reason that we are drawing QGraphicsItems and storing QPoint and QPolygonF? Would it make any sense
        to store the QGraphicsItems in fiducials or polylines?
     - Deep learning-based image super-resolution (or other approaches, see Eugene Myers', but also others').
     - Find a better initial image.
     - Include default, initial image in a data file.
     - Rescale intensity (for display, currently just linear stretch, but within skimage.exposure, there are multiple
     options for this. Also, in QtGui.QImage (see displayData) Format_Alpha8 gives you images a la Tony Harris, with inverted signal.
     - Enhance with Laplacian (see Fig. 3.40 in Gonzalez&Woods, 2nd ed).

     NOT SO IMPORTANT
     - Segment only in ROI (see call to watershed in rimcore, it has a mask parameter for this).
     - Go to time point.
     - Move to time point 1, 2, 3, 4, ...
     - Jump 10, 20, 30, 40, etc, time points.
     - Color images & split channels: how important is this? That is barely used in SIESTA.
     - Encode initial image, rather than distributing it as an image.


     DONE (in chronological order):
     - Load initial image.
     - Compare pillow.Image and scikit-image (which is a lot better documented). ENDED UP MOVING TO SCIKIT, BECAUSE
       OF THE BETTER DOC, BUT ALSO BECAUSE SO FAR WE ONLY USE LOAD METHODS, AND THOSE USE THE SAME LIBTIFF USED IN
       PILLOW. PLUS SCIKIT HAS A LOT MORE ALGORITHMIC DEVELOPMENT BEHIND.
     - Display coordinates and pixel value as mouse moves on image.
     - Zoom.
     - Polylines.
     - Delete all annotations in a time point/in the movie.
     - Cannot delete imported fiducials (but not problem with polylines).
     - drawPolyline/addPolyline and removePolyline need restructuring so that addPolyline does not depend on drawPolyline,
     but can be used to add a polyline to PyJAMAS in any time point (and thus called from loadannotations, importannotations,
     etc.
     - Show that PyJAMAS is faster than SIESTA handling images with LOTS of annotations.
     - Break event filter into smaller function calls. It's gonna become a monster function otherwise. Maybe even
     take it outside and create an eventListener class that does all the event handling.
     - Deleting a rectangle has some weird bug when deleting a rectangle overlapped by another one. From that
        point on, if you try to delete the last rectangle drawn, you will not see the rectangle disappear until
        the mouse is pressed again.
     - LiveWire.
     - Refactor rimcore to use row, col, not x, y (N4, N8, ND).
     - Profile with CProfile/
     - Recover focus after import SIESTA annotations.
     - requirements.txt (pipreqs .). Now one can install with pip install -r requirements.txt
     - Save annotations as xml files (similar to labelImg, to train TensorFlow. Or do csv directly, since xml
        annotations are converted into csv with python xml_to_csv.py
     - Still, some images will not open properly????? The pixel values are correct, but the order in whcih they are
        displayed is not. See, for instance, Clipboard_2.tif in the training folder. SOLVED BY ADDING AN ADDITIONAL PARAMETER
        IN THE QIMAGE CONSTRUCTOR USED IN DISPLAY_DATA.
     - When drawing polygons and rectangles, limit coordinates to within the image (or clip them at the end). Righjt now
        it is possible to draw beyond the image limits if the window is bigger than the image.
     - Measure polygons (skimage.measure). Need to check coordinate range and exit if polygon outside the image.
     - Brush size.
     - Native Menu Bar.
     - Save time series.
     - Crop image
     - Crop into a new window.
     - Move polylines.
     - Optic flow
     - Seed propagation
     - Scrollbar to move across time points.
     - Hide annotations.
     - Display fiducial ids.
     - Watershed (with Gaussian filtering BEFORE!!!)
     - Import/Export SIESTA annotations
     - Apply centerSeeds only to central, not peripheral seeds.
     - Remove magic numbers from rcbsegment.
     - Use concave hull of seeds to decide which ones are at the edges (instead of convex hull).
     - Gradient-based watershed -> this does not work very well (double edges?)
     - Add options to flip, rotate.
     - Export tracked polygons.
     - Argument parser for MIP.
     - Add progress information to project and concatenate.
     - Batch classifier hard_negative_mining? Necessary to wrap rimclassifier function? NO, as fit in rimclassifier
       already calls hard_negative_mining.
     - Time points in non-max suppression.
     - Move progress dialog to rutils (?) and add progress info, for instance in batch classifier. ENDED UP MOVING
     PROGRESS INFORMATION TO STATUS BAR.
     - Maximum suppression dial.
     - Classifier (a different GUI?).
     - Change flow field interpolation to linear, as done in rflow.m. The bivariate spline performs significantly worse ...
     - Make sure that optic flow does not push points too close to the image boundaries or outside (or at least has a
     flag that prevents that).
     - Open files dragged onto the window (both images and annotations).
     - Make sure that the default working folder (self.cwd) is being used every time a file dialog is open.
     - Add about option that displays the LICENSE file.
     - Test other watershed functions (e.g. opencv). CURRENT ONE IS BAD!!!
     - When displaying fiducial ids, a newly added fiducial should also get its id displayed.
     - Delete all seeds outside/inside polygon -> test skimage,measure.points_in_poly function to test polygon inclusion.
     - Use a different navigation toolbar (or no navigation toolbar) in matplotlibdialog.
     - Add zoom option to preview dialog in find seeds
     - Font size in preview window in findseedsdialog becomes smaller after the first live value change.
     - Invert image.
     - Segment objects localized by classifier: 1. Segmenting cardioblasts - for each detection, take the image, adaptive threshold, distance transform, identify one seed (close to the centre, discard image edges), identify a second seed (in the pixel with the lowest pixel value in the image - is this necessary?), and expand using the watershed on the gradient of the original image (or the inverted original image). Get rid of contours overlapping more than 10% with a pre-existing one when you are done (or as you go).
     - Find seeds - add polygon centroids as fiducials.
     - Track fiducials.
     - Autoplay movie (+ fps option)
     - Kymographs
     - Display polyline ids.
     - Track fiducials should extend to polylines.
     - Logistic regression.
     - Image registration.
     - Launch thread to do fiducial tracking, signal propagation, and display progress in the progress bar.
     - Add option to go to slice 0/end using RCBTimeSlicer.cbGoto.
     - Progress in statusbar is not visible!!! Is this because the dialogs are modal? Need multithreading!! See cbPlay, play_movie and progress_fn in rcbimage, and pjsthreads.
     - Fix error with export annotations to SIESTA format.
     - Copy&paste polylines.
      '''

    # Logo path.
    logo_filename = 'pyjamas.tif'
    logo_folder = os.path.split(__file__)[0]
    logo_path = os.path.join(logo_folder, logo_filename)

    # Annotation modes. @todo: convert into enum and use auto() to generate constants (?).
    no_annotations: int = 0
    fiducials: int = 1
    rectangles: int = 2
    polylines: int = 3
    livewire: int = 4
    move_polyline: int = 5
    export_fiducial_polyline: int = 6
    delete_fiducials_outside_polyline: int = 7
    delete_fiducials_inside_polyline: int = 8
    copy_polyline: int = 9

    # Data file extension.
    data_extension: str = '.pjs'
    matlab_extension: str = '.mat'
    image_extensions: Tuple[str] = ('.tif', '.tiff', '.jpg')

    # Plotting constants.
    fiducial_color = QtCore.Qt.red
    polyline_color = QtCore.Qt.green
    fiducial_radius: int = 6
    fiducial_brush_style: QtGui.QBrush = QtGui.QBrush()  # No fill for fiducials (open circles).
    polyline_brush_style: QtGui.QBrush = QtGui.QBrush()
    fiducial_font: QtGui.QFont = QtGui.QFont('Arial', 8)

    zoom_factors: tuple = (1., 2., 4., 8., .125, .25, .5)

    livewire_margin: int = 5

    def __init__(self):
        self.setupUI()  # Build the GUI.
        self.initData()  # Initialize object variables.

    def setupUI(self):
        self.app = QtWidgets.QApplication(sys.argv)

        import pyjamas.pjseventfilter as pjseventfilter
        import pyjamas.rcallbacks as rcallbacks

        QtCore.QObject.__init__(self)
        self.MainWindow = DragDropMainWindow()

        self.MainWindow.dropped.connect(self.file_dropped)

        self.MainWindow.setObjectName('PyJAMAS')
        self.MainWindow.resize(1183, 761)
        self.MainWindow.setWindowTitle('PyJAMAS')

        self.gScene = QtWidgets.QGraphicsScene(self.MainWindow)
        self.gScene.setSceneRect(0, 0, 1183, 761)
        self.gScene.setObjectName('gScene')

        self.gView = QtWidgets.QGraphicsView(self.gScene)
        self.gView.setObjectName('gView')
        self.gView.setMouseTracking(
            True)  # This here is necessary so that MouseMove events are triggered when no mouse buttons are pressed.
        self.filter = pjseventfilter.PJSEventFilter(self)
        self.gView.viewport().installEventFilter(self.filter)
        self.MainWindow.setCentralWidget(self.gView)

        self.timeSlider = QtWidgets.QSlider(self.MainWindow)
        self.timeSlider.setGeometry(QtCore.QRect(0, 0, 1183, 22))
        self.timeSlider.setOrientation(QtCore.Qt.Horizontal)
        self.timeSlider.setObjectName("timeSlider")
        self._cbTimeSlider = rcallbacks.rcbtimeslider.RCBTimeSlider(self)
        self.timeSlider.valueChanged.connect(self._cbTimeSlider.cbTimeSlider)

        self.menubar = QtWidgets.QMenuBar(self.MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1183, 22))
        self.menubar.setObjectName('menubar')
        self.menuIO = QtWidgets.QMenu(self.menubar)
        self.menuIO.setObjectName('menuIO')
        self.menuOptions = QtWidgets.QMenu(self.menubar)
        self.menuOptions.setObjectName('menuOptions')
        self.menuImage = QtWidgets.QMenu(self.menubar)
        self.menuImage.setObjectName('menuImage')
        self.menuClassifiers = QtWidgets.QMenu(self.menuImage)
        self.menuClassifiers.setObjectName('menuClassifiers')
        self.menuAnnotations = QtWidgets.QMenu(self.menubar)
        self.menuAnnotations.setObjectName('menuAnnotations')
        self.menuMeasurements = QtWidgets.QMenu(self.menubar)
        self.menuMeasurements.setObjectName('menuMeasurements')
        self.menuBatch = QtWidgets.QMenu(self.menubar)
        self.menuBatch.setObjectName('menuBatch')
        self.menuHelp = QtWidgets.QMenu(self.menubar)
        self.menuHelp.setObjectName('menuHelp')
        self.MainWindow.setMenuBar(self.menubar)

        self.statusbar = QtWidgets.QStatusBar(self.MainWindow)
        self.statusbar.setObjectName('statusbar')
        self.MainWindow.setStatusBar(self.statusbar)

        self.threadpool = QtCore.QThreadPool()
        #print("Multithreading with maximum %d threads" % self.threadpool.maxThreadCount())

        # IO MENU ----------------------------------------------------------------------
        # Load image action
        self._cbLoadTS = rcallbacks.rcbloadtimeseries.RCBLoadTimeSeries(self)
        self.addMenuItem(self.menuIO, 'Load grayscale image ...', QtCore.Qt.Key_T,
                         self._cbLoadTS.cbLoadTimeSeries)

        # Save time-lapse.
        self._cbSaveTS = rcallbacks.rcbsavetimeseries.RCBSaveTimeSeries(self)
        self.addMenuItem(self.menuIO, 'Save grayscale image ...', QtCore.Qt.Key_S,
                         self._cbSaveTS.cbSaveTimeSeries)

        # Save time-lapse, but only ROI inside the polygon.
        self._cbSaveSubim = rcallbacks.rcbsaveimage.RCBSaveImage(self)
        self.addMenuItem(self.menuIO, 'Save grayscale image in polygon ...', QtCore.Qt.Key_U,
                         partial(self._cbSaveSubim.cbSaveImage, filename=''))

        self.menuIO.addSeparator()

        # Load annotations.
        self._cbLoadAnn = rcallbacks.rcbloadannotations.RCBLoadAnnotations(self)
        self.addMenuItem(self.menuIO, 'Load annotations ...', QtCore.Qt.SHIFT + QtCore.Qt.Key_L,
                         self._cbLoadAnn.cbLoadAnnotations)

        # Save annotations.
        self._cbSaveAnn = rcallbacks.rcbsaveannotations.RCBSaveAnnotations(self)
        self.addMenuItem(self.menuIO, 'Save annotations ...', QtCore.Qt.Key_A,
                         self._cbSaveAnn.cbSaveAnnotations)

        self._cbExportPolylineAnn = rcallbacks.rcbexportpolylineannotations.RCBExportPolylineAnnotations(self)
        self.addMenuItem(self.menuIO, 'Export individual fiducial-polyline annotations ...', None,
                         self._cbExportPolylineAnn.cbExportPolylineAnnotations)

        self.menuIO.addSeparator()

        self._cbLoadClassifier = rcallbacks.rcbloadclassifier.RCBLoadClassifier(self)
        self.addMenuItem(self.menuIO, 'Load classifier ...', None,
                         self._cbLoadClassifier.cbLoadClassifier)

        self._cbSaveClassifier = rcallbacks.rcbsaveclassifier.RCBSaveClassifier(self)
        self.addMenuItem(self.menuIO, 'Save current classifier ...', None,
                         self._cbSaveClassifier.cbSaveClassifier)

        self.menuIO.addSeparator()

        # Save XML roi (for TensorFlow training).
        self._cbSaveAnnXML = rcallbacks.rcbsaveannotationsXML.RCBSaveAnnotationsXML(self)
        self.addMenuItem(self.menuIO, 'Save XML ROI ...', QtCore.Qt.Key_X,
                         self._cbSaveAnnXML.cbSaveAnnotationsXML)

        self.menuIO.addSeparator()

        # Import SIESTA annotations.
        self._cbImportSIESTAAnn = rcallbacks.rcbimportsiestaannotations.RCBImportSIESTAAnnotations(self)
        self.addMenuItem(self.menuIO, 'Import SIESTA annotations ...', None,
                         self._cbImportSIESTAAnn.cbImportSIESTAAnnotations)

        # Export SIESTA annotations.
        self._cbExportSIESTAAnn = rcallbacks.rcbexportsiestaannotations.RCBExportSIESTAAnnotations(self)
        self.addMenuItem(self.menuIO, 'Export SIESTA annotations ...', None,
                         self._cbExportSIESTAAnn.cbExportSIESTAAnnotations)

        self.menuIO.addSeparator()

        # Save display.
        self._cbSaveDisplay = rcallbacks.rcbsavedisplay.RCBSaveDisplay(self)
        self.addMenuItem(self.menuIO, 'Save display ...', None, self._cbSaveDisplay.cbSaveDisplay)

        self.menuIO.addSeparator()

        self._cbQuit = rcallbacks.rcbquit.RCBQuit(self)
        self.addMenuItem(self.menuIO, 'See ya\'!!', QtCore.Qt.ALT + QtCore.Qt.Key_Q, self._cbQuit.cbQuit)
        self.menubar.addMenu(self.menuIO)

        # OPTIONS MENU ----------------------------------------------------------------------
        # Set brush size.
        self._cbBrushSize = rcallbacks.rcbsetbrushsize.RCBSetBrushSize(self)
        self.addMenuItem(self.menuOptions, 'Set brush size ...', QtCore.Qt.Key_B,
                         self._cbBrushSize.cbSetBrushSize)

        self._cbFiducalIDs = rcallbacks.rcbdisplayfiducialids.RCBDisplayFiducialIDs(self)
        self.addMenuItem(self.menuOptions, 'Display fiducial and polyline ids', QtCore.Qt.Key_I,
                         self._cbFiducalIDs.cbDisplayFiducialIDs)

        self._cbFramesPerSec = rcallbacks.rcbframespersec.RCBFramesPerSec(self)
        self.addMenuItem(self.menuOptions, 'Set frames per sec ...', None,
                         self._cbFramesPerSec.cbFramesPerSec)

        self._cbCWD = rcallbacks.rcbsetcwd.RCBSetCWD(self)
        self.addMenuItem(self.menuOptions, 'Set working folder ...', None, self._cbCWD.cbSetCWD)

        self.menubar.addMenu(self.menuOptions)

        # IMAGE MENU ----------------------------------------------------------------------
        # Rotate image CW.
        self._cbImage = rcallbacks.rcbimage.RCBImage(self)
        self.addMenuItem(self.menuImage, 'Play sequence', QtCore.Qt.Key_Backslash, self._cbImage.cbPlay)

        self.menuImage.addSeparator()

        self.addMenuItem(self.menuImage, 'Rotate 90 degree clockwise', QtCore.Qt.SHIFT + QtCore.Qt.Key_Right,
                         partial(self._cbImage.cbRotateImage,
                                 direction=rcallbacks.rcbimage.RCBImage.CW))

        # Rotate image CCW.
        self.addMenuItem(self.menuImage, 'Rotate 90 degree counter-clockwise', QtCore.Qt.SHIFT + QtCore.Qt.Key_Left,
                         partial(self._cbImage.cbRotateImage,
                                 direction=rcallbacks.rcbimage.RCBImage.CCW))

        # Flip left-right.
        self.addMenuItem(self.menuImage, 'Flip left-right', QtCore.Qt.AltModifier + QtCore.Qt.Key_Right,
                         partial(self._cbImage.cbFlipImage,
                                 direction=rcallbacks.rcbimage.RCBImage.LEFT_RIGHT))

        # Flip top-bottom.
        self.addMenuItem(self.menuImage, 'Flip up-down', QtCore.Qt.AltModifier + QtCore.Qt.Key_Up,
                         partial(self._cbImage.cbFlipImage,
                                 direction=rcallbacks.rcbimage.RCBImage.UP_DOWN))

        self.menuImage.addSeparator()

        # Invert image.
        self.addMenuItem(self.menuImage, 'Invert image', None, self._cbImage.cbInvertImage)

        # Maximum intensity projection.
        self.addMenuItem(self.menuImage, 'Maximum intensity projection', None, self._cbImage.cbMIPImage)

        self.menuImage.addSeparator()

        self.addMenuItem(self.menuImage, 'Register image', QtCore.Qt.SHIFT + QtCore.Qt.Key_R, self._cbImage.cbRegisterImage)

        self.menuImage.addSeparator()
        # Zoom action.
        self._cbZoom = rcallbacks.rcbzoom.RCBZoom(self)
        self.addMenuItem(self.menuImage, 'Zoom', QtCore.Qt.Key_Z, self._cbZoom.cbZoom)

        # Crop action.
        self._cbCrop = rcallbacks.rcbcrop.RCBCrop(self)
        self.addMenuItem(self.menuImage, 'Crop', QtCore.Qt.SHIFT + QtCore.Qt.Key_P, self._cbCrop.cbCrop)

        # Crop action.
        self._cbCropNewW = rcallbacks.rcbcrop.RCBCrop(self)
        # Here I use partial to pass parameters to the cbCrop function.
        self.addMenuItem(self.menuImage, 'Crop into a new window', None, partial(self._cbCropNewW.cbCrop,
                                                                                 new_window=True))

        self._cbKymograph = rcallbacks.rcbkymograph.RCBKymograph(self)
        self.addMenuItem(self.menuImage, 'Kymograph', None, self._cbKymograph.cbKymograph)

        self.menuImage.addSeparator()

        # Find seeds.
        self._cbSegmentation = rcallbacks.rcbsegment.RCBSegment(self)
        self.addMenuItem(self.menuImage, 'Find seeds ...', QtCore.Qt.SHIFT + QtCore.Qt.Key_S,
                         self._cbSegmentation.cbFindSeeds)

        # Add fiducials in the centre of polylines.
        self.addMenuItem(self.menuImage, 'Add seeds in polyline centroids ...', QtCore.Qt.SHIFT + QtCore.Qt.Key_C,
                         self._cbSegmentation.cbCentroidSeeds)

        # Propagate seeds.
        self.addMenuItem(self.menuImage, 'Propagate seeds ...', QtCore.Qt.SHIFT + QtCore.Qt.Key_E,
                         self._cbSegmentation.cbPropagateSeeds)

        # Expand seeds.
        self.addMenuItem(self.menuImage, 'Expand seeds ...', QtCore.Qt.SHIFT + QtCore.Qt.Key_X,
                         self._cbSegmentation.cbExpandSeeds)

        # Expand and propagate seeds.
        self.addMenuItem(self.menuImage, 'Expand and propagate seeds ...', QtCore.Qt.SHIFT + QtCore.Qt.Key_Z,
                         self._cbSegmentation.cbExpandNPropagateSeeds)

        # Classifiers.
        self.menuImage.addMenu(self.menuClassifiers)
        # -- Display Classifier: YellowBrick?
        self._cbCreateLR = rcallbacks.rcbcreateLR.RCBCreateLR(self)
        self.addMenuItem(self.menuClassifiers, 'Create and train logistic regression model ...', None,
                         self._cbCreateLR.cbCreateLR)
        self._cbCreateSVM = rcallbacks.rcbcreateSVM.RCBCreateSVM(self)
        self.addMenuItem(self.menuClassifiers, 'Create and train support vector machine ...', None,
                         self._cbCreateSVM.cbCreateSVM)
        # -- Decision Trees

        self.menuClassifiers.addSeparator()

        self._cbApplyClassifier = rcallbacks.rcbapplyclassifier.RCBApplyClassifier(self)
        self.addMenuItem(self.menuClassifiers, 'Apply classifier ...', None, self._cbApplyClassifier.cbApplyClassifier)

        self.menuClassifiers.addSeparator()

        self._cbNonMaxSuppression = rcallbacks.rcbnonmax.RCBNonMax(self)
        self.addMenuItem(self.menuClassifiers, 'Non-maximum suppression ...', None,
                         self._cbNonMaxSuppression.cbNonMaxSuppression)

        self.addMenuItem(self.menuClassifiers, 'Segment detected objects ...', None, self._cbSegmentation.cbSegmentDetectedObjects)

        self.menuImage.addSeparator()

        # Next image action
        self._cbNextFrame = rcallbacks.rcbnextframe.RCBNextFrame(self)
        self.addMenuItem(self.menuImage, 'Next frame', QtCore.Qt.Key_Period, self._cbNextFrame.cbNextFrame)

        # Previous image action
        self._cbPrevFrame = rcallbacks.rcbprevframe.RCBPrevFrame(self)
        self.addMenuItem(self.menuImage, 'Previous frame', QtCore.Qt.Key_Comma, self._cbPrevFrame.cbPrevFrame)

        # Beginning and end actions
        self.addMenuItem(self.menuImage, 'Go to beginning', QtCore.Qt.Key_1, partial(self._cbTimeSlider.cbGoTo, slice_index=0))
        self.addMenuItem(self.menuImage, 'Go to end', QtCore.Qt.Key_0, partial(self._cbTimeSlider.cbGoTo, slice_index=-1))

        self.menubar.addMenu(self.menuImage)

        # ANNOTATIONS MENU ----------------------------------------------------------------------
        # No annotations.
        self._cbNoAnn = rcallbacks.rcbnoann.RCBNoAnn(self)
        self.addMenuItem(self.menuAnnotations, 'No annotations', QtCore.Qt.Key_N, self._cbNoAnn.cbNoAnn)

        self._cbHideAnn = rcallbacks.rcbhideann.RCBHideAnn(self)
        self.addMenuItem(self.menuAnnotations, 'Hide/display annotations', QtCore.Qt.Key_H, self._cbHideAnn.cbHideAnn)

        # Fiducials
        self._cbFidu = rcallbacks.rcbfiducials.RCBFiducials(self)
        self.addMenuItem(self.menuAnnotations, 'Fiducials', QtCore.Qt.Key_F, self._cbFidu.cbFiducials)

        # Rectangles
        self._cbRect = rcallbacks.rcbrectangles.RCBRectangles(self)
        self.addMenuItem(self.menuAnnotations, 'Rectangles', QtCore.Qt.Key_R, self._cbRect.cbRectangles)

        # Polylines
        self._cbPoly = rcallbacks.rcbpolylines.RCBPolylines(self)
        self.addMenuItem(self.menuAnnotations, 'Polylines', QtCore.Qt.Key_Y, self._cbPoly.cbPolylines)

        # Livewire
        self._cbLW = rcallbacks.rcblivewire.RCBLiveWire(self)
        self.addMenuItem(self.menuAnnotations, 'LiveWire', QtCore.Qt.Key_W, self._cbLW.cbLiveWire)

        self.menuAnnotations.addSeparator()

        # Copy and paste polyline
        self._cbCopyPastePoly = rcallbacks.rcbcopypastepolyline.RCBCopyPastePolyline(self)
        self.addMenuItem(self.menuAnnotations, 'Copy polyline', QtCore.Qt.Key_C, self._cbCopyPastePoly.cbCopyPolyline)

        self.addMenuItem(self.menuAnnotations, 'Paste polyline', QtCore.Qt.Key_V, self._cbCopyPastePoly.cbPastePolyline)

        # Move polyline.
        self._cbMovePoly = rcallbacks.rcbmovepolyline.RCBMovePolyline(self)
        self.addMenuItem(self.menuAnnotations, 'Move polyline', QtCore.Qt.Key_M, self._cbMovePoly.cbMovePolyline)

        self.menuAnnotations.addSeparator()

        self._cbTrackFidu = rcallbacks.rcbtrackfiducials.RCBTrackFiducials(self)
        self.addMenuItem(self.menuAnnotations, 'Track fiducials ...', QtCore.Qt.SHIFT + QtCore.Qt.Key_T,
                         self._cbTrackFidu.cbTrackFiducials)

        self.menuAnnotations.addSeparator()

        # Delete annotations on the current image.
        self._cbDeleteCurrentAnn = rcallbacks.rcbdeletecurrentann.RCBDeleteCurrentAnn(self)
        self.addMenuItem(self.menuAnnotations, 'Delete all annotations on current frame', QtCore.Qt.Key_Minus,
                         self._cbDeleteCurrentAnn.cbDeleteCurrentAnn)

        # Delete ALL annotations.
        self._cbDeleteAllAnn = rcallbacks.rcbdeleteallann.RCBDeleteAllAnn(self)
        self.addMenuItem(self.menuAnnotations, 'Delete all annotations in the sequence', QtCore.Qt.Key_Underscore,
                         self._cbDeleteAllAnn.cbDeleteAllAnn)

        # Delete fiducials OUTSIDE polyline.
        self._cbDeleteFiducialsOutsidePoly = rcallbacks.rcbdeletefiducialsinoutpoly.RCBDeleteFiducialsPoly(self)
        self.addMenuItem(self.menuAnnotations, 'Delete fiducials outside polyline',
                         QtCore.Qt.SHIFT + QtCore.Qt.Key_O,
                         self._cbDeleteFiducialsOutsidePoly.cbDeleteFiducialsOutsidePoly)

        self.menubar.addMenu(self.menuAnnotations)

        # Delete fiducials INSIDE polyline.
        self._cbDeleteFiducialsInsidePoly = rcallbacks.rcbdeletefiducialsinoutpoly.RCBDeleteFiducialsPoly(self)
        self.addMenuItem(self.menuAnnotations, 'Delete fiducials inside polyline',
                         QtCore.Qt.SHIFT + QtCore.Qt.Key_I,
                         self._cbDeleteFiducialsInsidePoly.cbDeleteFiducialsInsidePoly)

        self.menubar.addMenu(self.menuAnnotations)

        # MEASUREMENTS MENU --------------------------------------------------------------------
        self._cbMeasurePoly = rcallbacks.rcbmeasurepoly.RCBMeasurePoly(self)
        self.addMenuItem(self.menuMeasurements, 'Measure polylines ...', None, self._cbMeasurePoly.cbMeasurePoly)

        self.menubar.addMenu(self.menuMeasurements)

        # BATCH MENU ---------------------------------------------------------------------------
        # Project and concatenate.
        self._cbBatchProjectConcatenate = rcallbacks.rcbbatchprojectconcatenate.RCBBatchProjectConcat(self)
        self.addMenuItem(self.menuBatch, "Project and concatenate ...", None,
                         self._cbBatchProjectConcatenate.cbBatchProjectConcat)

        self.menubar.addMenu(self.menuBatch)

        # HELP MENU ---------------------------------------------------------------------------
        # Project and concatenate.
        self._cbAbout = rcallbacks.rcbabout.RCBAbout(self)
        self.addMenuItem(self.menuHelp, "About", None,
                         self._cbAbout.cbAbout)

        self.menubar.addMenu(self.menuHelp)

        self.menubar.setNativeMenuBar(True)
        self.retranslateUi(self.MainWindow)

    def initData(self):
        self._imageItem = None  # pixmap containing the image currently being displayed.
        self.zoom_index = 0  # index into PyJAMAS.zoom_factors to establish the zoom level.
        self.brush_size: int = 3  # brush size to paint polylines
        self.show_annotations: bool = True  # display or not annotations.
        self.display_fiducial_ids: bool = False  # display an identifier next to each fiducial
        self.fps: int = 7  # frames per second used to play the current sequence or when exporting as a movie.
        self.annotation_mode = PyJAMAS.no_annotations
        self.batch_classifier: BatchClassifier = None

        self._cbLoadTS.cbLoadTimeSeries(PyJAMAS.logo_path)
        self.cwd = os.getcwd()

        QtCore.QMetaObject.connectSlotsByName(self.MainWindow)

        self.gView.show()
        self.MainWindow.show()

        self._poly_ = []  # Stores polyline coordinates while they are being drawn. Dunder (__a__) is reserved for special methods (e.g. __len__()).
        self._copied_poly_ = None  # Stores copied polyline to be pasted.
        self._agraphicsitem_ = None  # Stores a graphicsitem transiently (e.g. a rectangle as it is being dragged).

        return True

    def prepare_image(self):
        if len(self.slices.shape) == 2:
            self.slices = numpy.expand_dims(self.slices, axis=0)

        self.curslice: int = 0
        self.imagedata: numpy.ndarray = self.slices[self.curslice]
        self.n_frames: int = self.slices.shape[0]
        self.height: int = self.slices.shape[1]  # number of rows.
        self.width:int = self.slices.shape[2]  # number of columns.

    def initImage(self):
        self.prepare_image()

        self.fiducials = [[] for i in range(self.n_frames)]
        self.polylines = [[] for i in range(self.n_frames)]

        # Make sure to continue to store the classifier in memory.
        if self.batch_classifier is not None:
            self.batch_classifier = BatchClassifier(self.n_frames, self.batch_classifier.image_classifier)
        else:
            self.batch_classifier = BatchClassifier(self.n_frames)

        self.displayData()

        # Resizing the window.
        self.gScene.setSceneRect(0, 0, self.width, self.height)
        self.MainWindow.resize(self.width * self.zoom_factors[self.zoom_index],
                               self.height * self.zoom_factors[self.zoom_index] + 60)
        self.timeSlider.setGeometry(QtCore.QRect(0, self.height + 18, self.MainWindow.width(), 22))
        self.timeSlider.setMinimum(1)
        self.timeSlider.setMaximum(self.n_frames)
        self.timeSlider.setValue(1)

    def retranslateUi(self, PyJAMAS):
        _translate = QtCore.QCoreApplication.translate
        self.menuIO.setTitle(_translate('PyJAMAS', 'IO'))
        self.menuOptions.setTitle(_translate('PyJAMAS', 'Options'))
        self.menuImage.setTitle(_translate('PyJAMAS', 'Image'))
        self.menuClassifiers.setTitle(_translate('PyJAMAS', 'Classifiers'))
        self.menuAnnotations.setTitle(_translate('PyJAMAS', 'Annotations'))
        self.menuMeasurements.setTitle(_translate('PyJAMAS', 'Measurements'))
        self.menuBatch.setTitle(_translate('PyJAMAS', 'Batch'))
        self.menuHelp.setTitle(_translate('PyJAMAS', 'Help'))
        self.menuHelp.setTitle(_translate('PyJAMAS', 'Help'))

        return True

    def addMenuItem(self, themenu, theitemname, theshortcut, thecallbackfunction):
        newAction = QtWidgets.QAction(self.MainWindow)
        newAction.setEnabled(True)
        newAction.setText(theitemname)
        newAction.setIconText(theitemname)
        newAction.setToolTip(theitemname)
        newAction.setObjectName('action' + theitemname)
        newAction.triggered.connect(thecallbackfunction)

        if theshortcut != None:
            newAction.setShortcut(theshortcut)

            # There are other possible values for the ShortcutContext in Qt.ShortcutContext. This is key for shortcuts to work when setNativeMenuBar(True) is used.
            newAction.setShortcutContext(QtCore.Qt.WidgetWithChildrenShortcut)
            newShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(theshortcut),
                                              self.MainWindow)
            newShortcut.activated.connect(thecallbackfunction)

        themenu.addAction(newAction)

        return True

    def displayData(self):
        # Delete annotations from the screen.
        self.eraseAnnotations()

        # Stretch the dynamic range of the image and then convert to 8 bits.
        # There is an autocontrast function in pillow, but it only works on 8 bit grayscale or color data.
        # And converting to 8 bit and then doing the stretch leads to information loss and pixelated images.
        # img_16bit_to_8bit = RImage(self.imagedata).stretch()
        # img_16bit_to_8bit = numpy.array(img_16bit_to_8bit, dtype=numpy.uint8)

        # themin = numpy.min(self.imagedata)
        # themax = numpy.max(self.imagedata)

        # Now we stretch the image for display. I tested skimage.exposure.rescale_intensity and my own stretch method
        # (implemented  following the code in dipimage). stretch took 419 us, vs 484 us for rescale_intensity.
        # img_16bit_to_8bit = exposure.rescale_intensity(self.imagedata, in_range=(themin, themax), out_range=(0, 255))

        img_16bit_to_8bit = rimutils.stretch(self.imagedata)
        img_16bit_to_8bit = numpy.array(img_16bit_to_8bit, dtype=numpy.uint8)

        # Display image. The fourth parameter in the QImage constructor is 1*self.width for grayscale images,
        # and 3*self.width for color images.
        qImg = QtGui.QImage(bytes(img_16bit_to_8bit), self.width, self.height, self.width,
                            QtGui.QImage.Format_Grayscale8)
        pixmap = QtGui.QPixmap.fromImage(qImg)

        if self._imageItem != None:
            self.gScene.removeItem(self._imageItem)

        self._imageItem = QtWidgets.QGraphicsPixmapItem(pixmap)
        self.gScene.addItem(self._imageItem)

        if self.show_annotations:
            self.paintAnnotations()

        self.statusbar.showMessage(str(self.curslice + 1) + '/' + str(self.n_frames))

        return True

    # Perhaps create a class RAnnotations, and a PyJAMAS object has an RAnnotations property, that contains fields fiducials, polylines, etc. Does it contain the GraphicsScene as well? Yes, it is the argument to the constructor.
    # Add and remove methods take on an additional argument, the current slice. Constants are taken from class properties (PyJAMAS.fiducial_color, etc.).
    def addFiducial(self, x, y, z, paint=True):
        """

        :param x:
        :param y:
        :param z:
        :param paint: when set to False, the screen will not be repainted. This helps when adding fiducials on a different thread (which is not allowed to modify the GUI - see https://doc.qt.io/qt-5/thread-basics.html#gui-thread-and-worker-thread)
        :return:
        """
        self.fiducials[z].append([x, y])
        # print('Fiducials on frame ' + str(self.curslice) + ': ' + self.fiducials[self.curslice].__str__())

        # Add ellipse at (0,0). Then move it to the right position. This is important so that scenePos() returns the proper coordinates for the item.
        # If you add here directly in the (x, y) coordinates, scenePos() returns [0, 0].
        if paint and z == self.curslice:
            theItem = self.gScene.addEllipse(0, 0, PyJAMAS.fiducial_radius, PyJAMAS.fiducial_radius,
                                             PyJAMAS.fiducial_color,
                                             PyJAMAS.fiducial_brush_style)
            theItem.setPos(x - PyJAMAS.fiducial_radius / 2, y - PyJAMAS.fiducial_radius / 2)

            # If fiducial ids are on display, repaint them.
            if self.display_fiducial_ids:
                self.repaint()

        return True

    def findGraphicItem(self, x, y, class_type,
                        radius=fiducial_radius):  # Substitute in removeFiducial, removePolygon and movepolyline.
        theItems = self.gScene.items(
            QtCore.QRectF(x - radius / 4, y - radius / 4, radius / 2, radius / 2))

        # Because there are layers, if a rectangle was drawn on the ellipse, the rectangle will come first. So we look for the ellipse.
        for theClickedItem in theItems:
            if type(theClickedItem) == class_type:
                return (theClickedItem)

        return None

    def find_clicked_polyline(self, x: int, y: int) -> int:
        self._agraphicsitem_ = self.findGraphicItem(x, y, QtWidgets.QGraphicsPolygonItem)

        if type(self._agraphicsitem_) == QtWidgets.QGraphicsPolygonItem:
            theClickedPolygon = self._agraphicsitem_.polygon()

            try:
                index = self.polylines[self.curslice].index(theClickedPolygon)
            except LookupError:
                return -1
            else:
                self._poly_ = self.polylines[self.curslice][index]
                return index
        else:
            return -1

    def removeFiducial(self, x, y, z):
        # Grab items within a small square around the click point.
        """theItems = self.gScene.items(
            QtCore.QRectF(x - PyJAMAS.fiducial_radius / 4, y - PyJAMAS.fiducial_radius / 4, PyJAMAS.fiducial_radius / 2,
                          PyJAMAS.fiducial_radius / 2))

        # Because there are layers, if a rectangle was drawn on the ellipse, the rectangle will come first. So we look for the ellipse.
        for theClickedItem in theItems:
            if type(theClickedItem) == QtWidgets.QGraphicsEllipseItem:
                break
        """

        theClickedItem = self.findGraphicItem(x, y, QtWidgets.QGraphicsEllipseItem)

        # If you found an ellipse:
        if type(theClickedItem) == QtWidgets.QGraphicsEllipseItem:
            # Get coordinates. Find item in fiducial list. If you can find it, delete it there and delete the item from the scene.
            pos = theClickedItem.scenePos()
            deleteCoords = [int(pos.x() + PyJAMAS.fiducial_radius / 2), int(pos.y() + PyJAMAS.fiducial_radius / 2)]

            if deleteCoords in self.fiducials[z]:
                self.fiducials[z].remove(deleteCoords)
                self.gScene.removeItem(theClickedItem)
                # print('Fiducials on frame ' + str(self.curslice) + ': ' + self.fiducials[self.curslice].__str__())

                # If fiducial ids are on display, repaint them.
                if self.display_fiducial_ids:
                    self.repaint()

        return True

    def removeFiducialsPolyline(self, polyline: QtGui.QPolygonF = None, inside_flag: bool = True):
        # Go through the list of fiducials.
        polyline_list = RUtils.qpolygonf2ndarray(polyline)

        inside_poly_flags: numpy.ndarray = points_in_poly(self.fiducials[self.curslice], polyline_list)

        # To avoid deleting fiducials from the list we are checking.
        fiducial_list = self.fiducials[self.curslice].copy()

        for thefiducial, is_inside in zip(fiducial_list, inside_poly_flags):
            if (is_inside and inside_flag) or ((not is_inside) and (not inside_flag)):
                self.removeFiducial(thefiducial[0], thefiducial[1], self.curslice)

        return True

    def drawRectangle(self, x0, y0, x1, y1):
        # print([x0,y0].__str__() + " " + [x1, y1].__str__())
        if x0 > x1:
            dummy = x1
            x1 = x0
            x0 = dummy

        if y0 > y1:
            dummy = y1
            y1 = y0
            y0 = dummy

        # print([x0,y0].__str__() + " " + [x1, y1].__str__())
        pen = QtGui.QPen()
        pen.setWidth(self.brush_size)
        pen.setColor(PyJAMAS.polyline_color)

        # thepolyline = QtGui.QPolygonF()

        theItem = self.gScene.addRect(x0, y0, x1 - x0, y1 - y0, pen, PyJAMAS.polyline_brush_style)

        return theItem  # , thepolyline

    def drawPath(self, coordinates):
        thepolyline = QtGui.QPolygonF()

        for i, thepoint in enumerate(coordinates):
            thepolyline.append(QtCore.QPointF(thepoint[0], thepoint[1]))

        thepath = QtGui.QPainterPath()
        thepath.addPolygon(thepolyline)

        pen = QtGui.QPen()
        pen.setWidth(self.brush_size)
        pen.setColor(PyJAMAS.polyline_color)

        theItem = self.gScene.addPath(thepath, pen, PyJAMAS.polyline_brush_style)

        return theItem, thepolyline

    def drawPolyline(self, aQPolygonF):
        pen = QtGui.QPen()
        pen.setWidth(self.brush_size)
        pen.setColor(PyJAMAS.polyline_color)

        theItem = self.gScene.addPolygon(aQPolygonF, pen, PyJAMAS.polyline_brush_style)

        return theItem

    def addPolyline(self, coordinates: list, z: int, paint: bool = True) -> bool:
        thepolyline = QtGui.QPolygonF()

        for thepoint in coordinates:
            thepolyline.append(QtCore.QPointF(thepoint[0], thepoint[1]))

        self.polylines[z].append(thepolyline)

        # Draw the polyline if it is being added to the right time point.
        if paint and z == self.curslice:
            self.drawPolyline(thepolyline)

            # If ids are on display, repaint them.
            if self.display_fiducial_ids:
                self.repaint()

        return True

    def replacePolyline(self, index: int, coordinates: list, z: int, paint: bool = True) -> bool:
        thepolyline = QtGui.QPolygonF()

        for thepoint in coordinates:
            thepolyline.append(QtCore.QPointF(thepoint[0], thepoint[1]))

        self.polylines[z][index] = thepolyline

        # Draw the polyline if it is being added to the right time point.
        if paint and z == self.curslice:
            self.drawPolyline(thepolyline)

            # If ids are on display, repaint them.
            if self.display_fiducial_ids:
                self.repaint()

        return True

    def removePolyline(self, x, y, z):
        # Grab items within a small square around the click point.
        """theItems = self.gScene.items(
            QtCore.QRectF(x - PyJAMAS.fiducial_radius / 4, y - PyJAMAS.fiducial_radius / 4, PyJAMAS.fiducial_radius / 2,
                          PyJAMAS.fiducial_radius / 2))

        # Because there are layers, if something was drawn on the polygon, that other thing will come first. So we look for the first polygon.
        for theClickedItem in theItems:
            if type(theClickedItem) == QtWidgets.QGraphicsPolygonItem:
                break
        """

        theClickedItem = self.findGraphicItem(x, y, QtWidgets.QGraphicsPolygonItem)

        # print("(" + theClickedItem.scenePos().x().__str__() + ", " + theClickedItem.scenePos().y().__str__() + ") " + theClickedItem.type().__str__())
        # If you clicked on a polyline (not only the edge, but anywhere inside as well):
        if type(theClickedItem) == QtWidgets.QGraphicsPolygonItem:
            theClickedPolygon = theClickedItem.polygon()

            try:
                index = self.polylines[z].index(theClickedPolygon)
            except LookupError:
                return False
            else:
                self.polylines[z].pop(index)
                self.gScene.removeItem(theClickedItem)

                # If ids are on display, repaint them.
                if self.display_fiducial_ids:
                    self.repaint()

                return True
        else:
            return False

    def eraseAnnotations(self):
        allItems = self.gScene.items()

        for i in allItems:
            if type(i) in [QtWidgets.QGraphicsEllipseItem, QtWidgets.QGraphicsPolygonItem, QtWidgets.QGraphicsTextItem]:
                self.gScene.removeItem(i)

        return True

    def paintAnnotations(self):
        pen = QtGui.QPen()
        pen.setWidth(self.brush_size)
        pen.setColor(PyJAMAS.polyline_color)

        for i, thepoly in enumerate(self.polylines[self.curslice]):
            if thepoly == [] or list(thepoly) == []:
                continue

            self.gScene.addPolygon(thepoly, pen, PyJAMAS.polyline_brush_style)

            if self.display_fiducial_ids:
                polygon = RUtils.qpolygonf2polygon(thepoly)
                theItem = self.gScene.addText(str(i + 1), PyJAMAS.fiducial_font)
                theItem.setPos(polygon.centroid.x, polygon.centroid.y)
                theItem.setDefaultTextColor(self.polyline_color)

        # Paint fiducials after polylines so that polylines are in the foreground.
        for i, thefiducial in enumerate(self.fiducials[self.curslice]):
            # Add ellipse at (0,0). Then move it to the right position. This is important so that scenePos() returns the proper coordinates for the item.
            # If you add here directly in the (x, y) coordinates, scenePos() returns [0, 0].
            theItem = self.gScene.addEllipse(0, 0, self.fiducial_radius, self.fiducial_radius, self.fiducial_color,
                                             self.fiducial_brush_style)

            x = thefiducial[0]
            y = thefiducial[1]
            theItem.setPos(x - PyJAMAS.fiducial_radius / 2, y - PyJAMAS.fiducial_radius / 2)

            if self.display_fiducial_ids:
                theItem = self.gScene.addText(str(i + 1), PyJAMAS.fiducial_font)
                theItem.setPos(x - PyJAMAS.fiducial_radius / 2, y - PyJAMAS.fiducial_radius / 2)
                theItem.setDefaultTextColor(self.fiducial_color)

        # Make sure focus goes back to the main window.
        self.MainWindow.activateWindow()

        return True

    def repaint(self):
        self.eraseAnnotations()
        self.paintAnnotations()

        return True

    def file_dropped(self, l: list) -> bool:
        for file_name in l:
            if os.path.exists(file_name):
                file_name_str = str(file_name).lower()
                _, extension = os.path.splitext(file_name_str)

                if extension in rimcore.rimage.image_extensions:
                    self._cbLoadTS.cbLoadTimeSeries(file_name_str)
                elif extension == PyJAMAS.data_extension:
                    self._cbLoadAnn.cbLoadAnnotations(file_name_str)
                elif extension == PyJAMAS.matlab_extension:
                    self._cbImportSIESTAAnn.cbImportSIESTAAnnotations(file_name_str)

            else:
                return False

        return True

    def __copy__(self):
        newone = type(self)()
        newone.__dict__.update(self.__dict__)
        return newone

    def __deepcopy__(self, memodict={}):
        newone = type(self)()
        memodict[id(self)] = newone

        # Deep copy methods with a __deepcopy__ magic method.
        # Otherwise, copy by reference.
        for k, v in self.__dict__.items():
            if getattr(v, "__deepcopy__", None):
                setattr(newone, k, v.__deepcopy__(memodict))
            else:
                setattr(newone, k, v)

        return newone


def main():
    #app = QtWidgets.QApplication(sys.argv)

    aPyJAMA: PyJAMAS = PyJAMAS()
    sys.exit(aPyJAMA.app.exec_())


if __name__ == '__main__':
    main()



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gzip
import os
import pickle

from PyQt5 import QtWidgets

import pyjamas.pjscore as pjscore
from .rcallback import RCallback


class RCBSaveAnnotations(RCallback):
    def cbSaveAnnotations(self, filename: str = '', polylines: list = None, fiducials: list = None):  # Handle IO errors.
        # Get file name.
        if filename == '' or filename is False: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            defaultname, _ = os.path.splitext(os.path.basename(self.pjs.filename))
            fname = QtWidgets.QFileDialog.getSaveFileName(None, 'Save annotations ...', self.pjs.cwd + os.sep + defaultname,
                                                          filter='PyJAMAS data (*' + pjscore.PyJAMAS.data_extension + ')')  # fname[0] is the full filename, fname[1] is the filter used.

            # If cancel ...
            if fname == '':
                return False

            # else ...
            else:
                filename = fname[0]

        self.pjs.cwd = os.path.dirname(filename)

        if filename[-4:] != pjscore.PyJAMAS.data_extension:
            filename = filename + pjscore.PyJAMAS.data_extension

        if not polylines:
            polylines = self.pjs.polylines

        if not fiducials:
            fiducials = self.pjs.fiducials

        # Prepare polygons to be stored: pickle does not support QPolygonF, so we convert the polygons to a list.
        # We store the polygons as QPolygonF because QGraphicsScene does have an addPolygon method that takes a
        # QPolygonF as the parameter.
        polyline_list = [[] for iframe in polylines]
        for iframe, theframepolylines in enumerate(polylines):
            for ipoly, thepolyline in enumerate(theframepolylines):
                polyline_list[iframe].append([])
                for ipnt, thepoint in enumerate(thepolyline):
                    polyline_list[iframe][ipoly].append([thepoint.x(), thepoint.y()])

        # Open file for writing.
        fh = None

        try:
            fh = gzip.open(filename, "wb")
            pickle.dump(fiducials, fh, pickle.HIGHEST_PROTOCOL)
            pickle.dump(polyline_list, fh, pickle.HIGHEST_PROTOCOL)

        except (IOError, OSError) as ex:
            if fh is not None:
                fh.close()

            print(ex)

        self.pjs.statusbar.showMessage(f"Saved {filename}.")

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt5 import QtWidgets

import pyjamas.dialogs as dialogs
from .rcallback import RCallback
import pyjamas.rimage.rimclassifier.rimlr as rimlr


class RCBCreateLR(RCallback):
    def cbCreateLR(self, parameters: dict = None) -> bool:  # Handle IO errors.
        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.logregression.LRDialog()
            ui.setupUi(dialog)

            dialog.exec_()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.Accepted
            parameters = ui.parameters()

            dialog.close()

        if continue_flag:
            self.pjs.batch_classifier.image_classifier = rimlr.lr(parameters)
            self.pjs.batch_classifier.fit()
            self.pjs.statusbar.showMessage("Trained logistic regression model!")  # todo: launch in thread.
            return True

        else:
            return False




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from sklearn.linear_model import LogisticRegression

from pyjamas.rimage.rimclassifier.rimclassifier import rimclassifier
from pyjamas.rutils import RUtils


class lr(rimclassifier):
    # Missed class penalty: small values (0.05) result in some additional nuclei detected and others
    # not. Large values (100) do not seem to change the result with respect to 1.0.
    DEFAULT_C: float = 1.0

    # Should featurecalculator be a class attribute?
    def __init__(self, parameters: dict = None):
        super().__init__(parameters)

        # SVM-specific parameters.
        self.misclass_penalty_C: float = parameters.get('C', lr.DEFAULT_C)
        self.classifier = parameters.get('classifier', LogisticRegression(C=self.misclass_penalty_C,
                                                                          solver='liblinear', random_state=1))

    def save_classifier(self, filename: str) -> bool:
        theclassifier = {
            'positive_training_folder': self.positive_training_folder,
            'negative_training_folder': self.negative_training_folder,
            'hard_negative_training_folder': self.hard_negative_training_folder,
            'train_image_size': self.train_image_size,
            'fc': self.fc,
            'step_sz': self.step_sz,
            'C': self.misclass_penalty_C,
            'iou_threshold': self.iou_threshold,
            'prob_threshold': self.prob_threshold,
            'max_num_objects_dial': self.max_num_objects,
            'classifier': self.classifier,
            'features_positive_array': self.features_positive_array,
            'features_negative_array': self.features_negative_array,
        }

        return RUtils.pickle_this(theclassifier, RUtils.set_extension(filename, rimclassifier.CLASSIFIER_EXTENSION))



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

from PyQt5 import QtCore, QtWidgets

import pyjamas.pjscore as pjscore
from .rcallback import RCallback


class RCBExportPolylineAnnotations(RCallback):
    FILENAME_BASE = "cell_"
    FILENAME_FIDUCIAL_LENGTH = 5

    def cbExportPolylineAnnotations(self, folder_name: str = '') -> bool:
        if folder_name != '' and folder_name is not False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            self.pjs.cwd = folder_name

        self.pjs.annotation_mode = pjscore.PyJAMAS.export_fiducial_polyline
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CrossCursor)

        return True

    def export_polyline_annotations(self, x: int, y: int) -> bool:
        # Make sure the fiducial is in the list.
        assert [x, y] in self.pjs.fiducials[self.pjs.curslice], "Fiducial not found!"

        # Identify the fiducial.
        fiducial_index: int = self.pjs.fiducials[self.pjs.curslice].index([x, y])

        # List of polygons and fiducials for new annotation file.
        fiducial_list: list = [[] for i in range(self.pjs.n_frames)]
        polyline_list: list = [[] for i in range(self.pjs.n_frames)]

        # In every image:
        for slice in range(self.pjs.n_frames):
            # If there are enough fiducials and any polylines.
            if self.pjs.fiducials[slice] and fiducial_index < len(self.pjs.fiducials[slice]) and self.pjs.polylines[slice]:
                # Store the fiducial.
                fiducial_list[slice].append(self.pjs.fiducials[slice][fiducial_index])

                # Find the first polygon that contains the fiducial and store it as well.
                thefiducial: list = self.pjs.fiducials[slice][fiducial_index]
                thepolylines: list = [one_polyline for one_polyline in self.pjs.polylines[slice]]

                for index_polyline, one_polyline in enumerate(thepolylines):
                    if one_polyline.containsPoint(QtCore.QPointF(thefiducial[0], thefiducial[1]), QtCore.Qt.OddEvenFill):
                        polyline_list[slice].append(self.pjs.polylines[slice][index_polyline])
                        break

        # Save new annotation file.
        filename: str = os.path.join(self.pjs.cwd, self.FILENAME_BASE + str(fiducial_index+1).zfill(self.FILENAME_FIDUCIAL_LENGTH))
        self.pjs._cbSaveAnn.cbSaveAnnotations(filename, polyline_list, fiducial_list)

        return True








"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
from PyQt5 import QtCore, QtWidgets

from pyjamas.pjscore import PyJAMAS
import pyjamas.rimage as rimage


class PJSEventFilter(QtCore.QObject):
    def __init__(self, ui):
        """

        :type ui: PyJAMAS
        """
        super().__init__()

        self.pjs: PyJAMAS = ui
        self.x_prev: int = -1
        self.y_prev: int = -1
        self.x: int = -1
        self.y: int = -1

    def eventFilter(self, source, event: QtCore.QEvent):
        # Event coordinates.
        if event.type() in (QtCore.QEvent.MouseMove, QtCore.QEvent.MouseButtonPress, QtCore.QEvent.MouseButtonRelease):
            thepoint = self.pjs.gView.mapToScene(event.x(), event.y())

            self.x, self.y = (int(numpy.round(thepoint.x())), int(numpy.round(thepoint.y())))

            # Check for boundaries.
            if self.x < 0:
                self.x = 0
            elif self.x >= self.pjs.width:
                self.x = self.pjs.width - 1

            if self.y < 0:
                self.y = 0
            elif self.y >= self.pjs.height:
                self.y = self.pjs.height - 1

        # Mouse click: display pixel coordinates and value.
        if event.type() == QtCore.QEvent.MouseMove:
            if 0 <= self.x < self.pjs.width and 0 <= self.y < self.pjs.height:
                self.pjs.statusbar.showMessage(
                    str(self.pjs.curslice + 1) + '/' + str(self.pjs.n_frames) + '\t(' + str(self.x) + ', ' + str(
                        self.y) + '): ' + str(self.pjs.imagedata[self.y, self.x]))

        # Fiducials: add and remove.
        if self.pjs.annotation_mode == PyJAMAS.fiducials and event.type() == QtCore.QEvent.MouseButtonPress:
            if event.buttons() == QtCore.Qt.LeftButton:
                self.pjs.addFiducial(self.x, self.y, self.pjs.curslice)

            elif event.buttons() == QtCore.Qt.RightButton:
                self.pjs.removeFiducial(self.x, self.y, self.pjs.curslice)

        # Rectangles.
        elif self.pjs.annotation_mode == PyJAMAS.rectangles:
            if event.type() == QtCore.QEvent.MouseButtonPress:
                # Store first coordinate if left click ...
                if event.buttons() == QtCore.Qt.LeftButton:
                    self.pjs._poly_ = [self.x, self.y]

                # ... or delete polygon if right click.
                elif event.buttons() == QtCore.Qt.RightButton:
                    self.pjs.removePolyline(self.x, self.y,
                                            self.pjs.curslice)  # CAN WE JUST HAVE ONE removeAnnotationItem function?

            # Redraw rectangle as the user drags the cursor.
            elif event.type() == QtCore.QEvent.MouseMove and event.buttons() == QtCore.Qt.LeftButton:  # Mouse move events will occur only when a mouse button is pressed down, unless mouse tracking has been enabled with QWidget.setMouseTracking()
                if self.pjs._agraphicsitem_:
                    self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                self.pjs._agraphicsitem_ = self.pjs.drawRectangle(self.pjs._poly_[0], self.pjs._poly_[1], self.x, self.y)
                self.pjs.statusbar.showMessage(
                    '{0}/{1}\t({2}, {3}) -> w: {4}, h: {5}'.format(str(self.pjs.curslice + 1),
                                                                                       str(self.pjs.n_frames), str(
                            self.pjs._poly_[0]), str(
                            self.pjs._poly_[1]), str(abs(self.pjs._poly_[0] - self.x)+1), str(abs(self.pjs._poly_[1] - self.y)+1)))

            # Add rectangle as mouse button is released.
            elif event.type() == QtCore.QEvent.MouseButtonRelease and self.pjs._poly_ != []:
                if self.pjs._agraphicsitem_:
                    self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                x0 = self.pjs._poly_[0]
                y0 = self.pjs._poly_[1]
                x1 = self.x
                y1 = self.y

                if x0 > x1:
                    dummy = x1
                    x1 = x0
                    x0 = dummy

                if y0 > y1:
                    dummy = y1
                    y1 = y0
                    y0 = dummy

                self.pjs.addPolyline([[x0, y0], [x0, y1], [x1, y1], [x1, y0]], self.pjs.curslice)
                self.pjs._poly_ = []

        # Polylines.
        elif self.pjs.annotation_mode == PyJAMAS.polylines:
            if event.type() == QtCore.QEvent.MouseButtonPress:
                # Store coordinate if left click ...
                if event.buttons() == QtCore.Qt.LeftButton:
                    self.pjs._poly_.append([self.x, self.y])

                    if self.pjs._agraphicsitem_:
                        self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                    self.pjs._agraphicsitem_, _ = self.pjs.drawPath(self.pjs._poly_)

                # Delete last point if middle click
                elif event.buttons() == QtCore.Qt.MiddleButton and self.pjs._poly_ is not []:
                    self.pjs._poly_ = self.pjs._poly_[0:-1]
                    if self.pjs._agraphicsitem_:
                        self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                        apoly = self.pjs._poly_.copy()
                        apoly.append([self.x, self.y])

                        self.pjs._agraphicsitem_, _ = self.pjs.drawPath(apoly)
                # ... or delete polygon if right click.
                elif event.buttons() == QtCore.Qt.RightButton:
                    self.pjs.removePolyline(self.x, self.y, self.pjs.curslice)

            # Redraw polyline as the user moves the cursor.
            elif event.type() == QtCore.QEvent.MouseMove and self.pjs._poly_ != []:
                if self.pjs._agraphicsitem_:
                    self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                    apoly = self.pjs._poly_.copy()
                    apoly.append([self.x, self.y])

                    self.pjs._agraphicsitem_, _ = self.pjs.drawPath(apoly)

            # Add polyline as mouse button is released.
            elif event.type() == QtCore.QEvent.MouseButtonDblClick and self.pjs._poly_ != []:
                if self.pjs._agraphicsitem_:
                    self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                self.pjs.addPolyline(self.pjs._poly_, self.pjs.curslice)
                self.pjs._poly_ = []

        # Delete fiducials outside polyline.
        elif self.pjs.annotation_mode == PyJAMAS.delete_fiducials_outside_polyline:
            if event.type() == QtCore.QEvent.MouseButtonPress:
                index = self.pjs.find_clicked_polyline(self.x, self.y)
                if index >= 0:
                    # Delete every annotation outside.
                    self.pjs.removeFiducialsPolyline(self.pjs._poly_, False)
                else:
                    return None

        # Delete fiducials inside polyline.
        elif self.pjs.annotation_mode == PyJAMAS.delete_fiducials_inside_polyline:
            if event.type() == QtCore.QEvent.MouseButtonPress:
                index = self.pjs.find_clicked_polyline(self.x, self.y)
                if index >= 0:
                    # Delete every annotation outside.
                    self.pjs.removeFiducialsPolyline(self.pjs._poly_, True)
                else:
                    return None

        # Copy polyline.
        elif self.pjs.annotation_mode == PyJAMAS.copy_polyline:
            if event.type() == QtCore.QEvent.MouseButtonPress:
                index = self.pjs.find_clicked_polyline(self.x, self.y)

                # Not using self.pjs._poly_ here, as other operations on the UI will overwrite it.
                self.pjs._copied_poly_ = self.pjs.polylines[self.pjs.curslice][index]

                if index < 0:
                    return None

        # LiveWire.
        elif self.pjs.annotation_mode == PyJAMAS.livewire:
            if event.type() == QtCore.QEvent.MouseButtonPress:
                # Store coordinate if left click ...
                if event.buttons() == QtCore.Qt.LeftButton:
                    if self.pjs._poly_ != []:
                        thesource = self.pjs._poly_[-1]
                        thedest = [self.x, self.y]
                        thepoints = rimage.rimcore.rimage(self.pjs.imagedata).livewire(thesource, thedest,
                                                                                       PyJAMAS.livewire_margin)  # Make livewire a static method?
                        self.pjs._poly_.extend(
                            thepoints)  ### CHANGE: STORE ALL COORDINATES GENERATED BY LIVEWIRE FROM LAST CLICKED POINT

                        if self.pjs._agraphicsitem_:
                            self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                        self.pjs._agraphicsitem_, _ = self.pjs.drawPath(self.pjs._poly_)

                    else:
                        self.pjs._poly_.append([self.x, self.y])
                        if self.pjs._agraphicsitem_:
                            self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                        self.pjs._agraphicsitem_, _ = self.pjs.drawPath(self.pjs._poly_)
                # Delete last point if middle click # NOT DONE YET.
                elif event.buttons() == QtCore.Qt.MiddleButton and self.pjs._poly_ is not []:
                    self.pjs._poly_ = self.pjs._poly_[0:-1]
                    if self.pjs._agraphicsitem_:
                        self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                        apoly = self.pjs._poly_.copy()
                        apoly.append([self.x, self.y])

                        self.pjs._agraphicsitem_, _ = self.pjs.drawPath(apoly)
                # ... or delete polygon if right click.
                elif event.buttons() == QtCore.Qt.RightButton:
                    self.pjs.removePolyline(self.x, self.y, self.pjs.curslice)

            # Redraw polyline as the user moves the cursor.
            elif event.type() == QtCore.QEvent.MouseMove and self.pjs._poly_ != []:
                if self.pjs._agraphicsitem_:
                    self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                    apoly = self.pjs._poly_.copy()
                    thesource = apoly.copy()[-1]
                    thedest = [self.x, self.y]
                    thepoints = rimage.rimcore.rimage(self.pjs.imagedata).livewire(thesource, thedest,
                                                                                   PyJAMAS.livewire_margin)  # Make livewire a static method? Or make self.imagedata an RImage?
                    apoly.extend(
                        thepoints)  ### CHANGE: STORE ALL COORDINATES GENERATED BY LIVEWIRE FROM LAST CLICKED POINT

                    self.pjs._agraphicsitem_, _ = self.pjs.drawPath(apoly)
                    # self.statusbar.showMessage(str(self.curslice + 1) + '/' + str(self.n_frames) + '\t(' + str(self._poly_[0]) + ', ' + str(
                    # self._poly_[1]) + ') -> w: ' + str(abs(self._poly_[0] - self.x)) + ', h: ' + str(abs(self._poly_[1] - self.y)))

            # Add polyline as mouse button is released.
            elif event.type() == QtCore.QEvent.MouseButtonDblClick and self.pjs._poly_ != []:
                if self.pjs._agraphicsitem_:
                    self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                thesource = self.pjs._poly_[-1]
                thedest = self.pjs._poly_[0]
                thepoints = rimage.rimcore.rimage(self.pjs.imagedata).livewire(thesource, thedest,
                                                                               PyJAMAS.livewire_margin)  # Make livewire a static method? Or make self.imagedata an RImage?
                self.pjs._poly_.extend(thepoints)
                self.pjs.addPolyline(self.pjs._poly_, self.pjs.curslice)
                self.pjs._poly_ = []

        # Move polylines.
        elif self.pjs.annotation_mode == PyJAMAS.move_polyline:
            if event.type() == QtCore.QEvent.MouseButtonPress:
                index = self.pjs.find_clicked_polyline(self.x, self.y)

                if index < 0:
                    return None

            # Redraw polyline as the user moves the cursor.
            elif event.type() == QtCore.QEvent.MouseMove and event.buttons() == QtCore.Qt.LeftButton and self.x_prev > -1 and self.y_prev > -1:
                if self.pjs._agraphicsitem_:
                    self.pjs.gScene.removeItem(self.pjs._agraphicsitem_)

                    theshift = QtCore.QPointF(self.x - self.x_prev, self.y - self.y_prev)

                    for i in range(self.pjs._poly_.size()):
                        self.pjs._poly_[i] += theshift

                    # Make sure you stay within the image: first find bounding box.
                    theboundingrect = self.pjs._poly_.boundingRect()

                    thex = theboundingrect.x()
                    they = theboundingrect.y()
                    themaxx = thex + theboundingrect.width()
                    themaxy = they + theboundingrect.height()

                    # Make sure the bounding box is within the image. If not, ...
                    if thex < 0 or they < 0 or themaxx >= self.pjs.width or themaxy >= self.pjs.height:
                        theshift = QtCore.QPointF(0, 0)

                        # Shift the X and/or Y coordinates accordingly.
                        if thex < 0:
                            theshift.setX(theshift.x() - thex)
                        if themaxx >= self.pjs.width:
                            theshift.setX(theshift.x() - (themaxx - self.pjs.width + 1))
                        if they < 0:
                            theshift.setY(theshift.y() - they)
                        if themaxy >= self.pjs.height:
                            theshift.setY(theshift.y() - (themaxy - self.pjs.height + 1))

                        # This is the shift happening!
                        for i in range(self.pjs._poly_.size()):
                            self.pjs._poly_[i] += theshift

                    self.pjs._agraphicsitem_ = self.pjs.drawPolyline(self.pjs._poly_)
                    self.pjs.repaint()

            # Update polyline as mouse button is released.
            elif event.type() == QtCore.QEvent.MouseButtonRelease and self.pjs._poly_ != []:
                self.pjs._poly_ = []

        # Export polylines associated with a fiducial.
        elif self.pjs.annotation_mode == PyJAMAS.export_fiducial_polyline:
            self.process_export_fiducial_polyline(event)

        elif event.type() == QtCore.QEvent.Resize:
            self.pjs.timeSlider.setGeometry(
                QtCore.QRect(0, self.pjs.MainWindow.height() - 43, self.pjs.MainWindow.width(), 22))

        # Store current position as previous (for mouse tracking).
        if event.type() in (QtCore.QEvent.MouseMove, QtCore.QEvent.MouseButtonPress, QtCore.QEvent.MouseButtonRelease):
            self.x_prev = self.x
            self.y_prev = self.y

        return QtCore.QObject.eventFilter(self.pjs, source, event)

    def process_export_fiducial_polyline(self, event):
        if event.type() == QtCore.QEvent.MouseButtonPress:
            theClickedItem = self.pjs.findGraphicItem(self.x, self.y, QtWidgets.QGraphicsEllipseItem)

            # If you found an ellipse:
            if type(theClickedItem) == QtWidgets.QGraphicsEllipseItem:
                # Get coordinates.
                pos = theClickedItem.scenePos()
                fiducial_coords = [int(pos.x() + PyJAMAS.fiducial_radius / 2),
                                   int(pos.y() + PyJAMAS.fiducial_radius / 2)]

                self.pjs._cbExportPolylineAnn.export_polyline_annotations(fiducial_coords[0], fiducial_coords[1])
        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
from PyQt5 import QtWidgets

import pyjamas.dialogs as dialogs
from pyjamas.pjsthreads import ThreadSignals
from pyjamas.rcallbacks.rcallback import RCallback


class RCBApplyClassifier(RCallback):
    def cbApplyClassifier(self, firstSlice: int = None, lastSlice: int = None) -> bool:  # Handle IO errors.
        if (firstSlice is False or firstSlice is None or lastSlice is False or lastSlice is None) and self.pjs is not None:
            dialog = QtWidgets.QDialog()
            ui = dialogs.timepoints.TimePointsDialog()

            if self.pjs.n_frames == 1:
                lastSlice = 1
            else:
                lastSlice = self.pjs.slices.shape[0]

            ui.setupUi(dialog, firstslice=self.pjs.curslice + 1, lastslice=lastSlice)

            dialog.exec_()
            dialog.show()
            # If the dialog was closed by pressing OK, then run the measurements.
            continue_flag = dialog.result() == QtWidgets.QDialog.Accepted
            firstSlice, lastSlice = ui.parameters()

            dialog.close()
        else:
            continue_flag = True

        if continue_flag:
            if firstSlice <= lastSlice:
                theslicenumbers = numpy.arange(firstSlice - 1, lastSlice, dtype=int)
            else:
                theslicenumbers = numpy.arange(lastSlice - 1, firstSlice, dtype=int)

            self.launch_thread(self.apply_classifier, {'theslices': theslicenumbers, 'progress': True, 'stop': True},
                               finished_fn=self.finished_fn,  progress_fn=self.progress_fn, stop_fn=self.stop_fn)

            return True
        else:
            return False

    def apply_classifier(self, theslices: numpy.ndarray, progress_signal: ThreadSignals, stop_signal: ThreadSignals) -> bool:
        # Make sure that the slices are in a 1D numpy array.
        theslices = numpy.atleast_1d(theslices)
        num_slices = theslices.size

        stop_signal.emit("Applying classifier ...")

        self.pjs.batch_classifier.predict(self.pjs.slices, theslices, progress_signal)

        # For every slice ...
        for index in theslices:
            self.add_classifier_boxes(self.pjs.batch_classifier.box_arrays[index], index)

        return True

    def add_classifier_boxes(self, boxes: numpy.ndarray = None, slice_index: int = None) -> bool:  # The first slice_index should be 0.
        if boxes is None or boxes is False or boxes == []:
            return False

        if slice_index is None or slice_index is False:
            slice_index = self.pjs.curslice

        for aBox in boxes:
            # Boxes stored as [minrow, mincol, maxrow, maxcol]
            self.pjs.addPolyline([[aBox[1], aBox[0]], [aBox[3], aBox[0]], [aBox[3], aBox[2]],
                                  [aBox[1], aBox[2]]], slice_index, paint=False)

        return True





"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .rcallback import RCallback
from PyQt5 import QtWidgets
from pyjamas.external import pascal_voc_io
import os


class RCBSaveAnnotationsXML(RCallback):
    def cbSaveAnnotationsXML(self, filename=''):  # Handle IO errors.
        # Get file name.
        fname = ''

        if filename == '' or filename is False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            defaultname, _ = os.path.splitext(os.path.basename(self.pjs.filename))
            fname = QtWidgets.QFileDialog.getSaveFileName(None, 'Save annotations ...',
                                                          self.pjs.cwd + os.sep + defaultname,
                                                          filter='XML file (*' + pascal_voc_io.XML_EXT + ')')  # fname[0] is the full filename, fname[1] is the filter used.

        # If cancel ...
        if fname == '':
            return False

        # else ...
        else:
            filename = fname[0]
            self.pjs.cwd = os.path.dirname(filename)

        if filename[-4:] != pascal_voc_io.XML_EXT:
            filename = filename + pascal_voc_io.XML_EXT

        writer = pascal_voc_io.PascalVocWriter(os.path.split(self.pjs.cwd)[1], os.path.basename(self.pjs.filename),
                                               [self.pjs.height, self.pjs.width, 1], localImgPath=self.pjs.filename)
        writer.verified = False

        theframepolylines = self.pjs.polylines[self.pjs.curslice]
        for ipoly, thepolyline in enumerate(theframepolylines):
            points = []

            for ipnt, thepoint in enumerate(thepolyline):
                points.append([thepoint.x(), thepoint.y()])

            label = 'cell'
            difficult = 0
            bndbox = RCBSaveAnnotationsXML.convertPoints2BndBox(points)

            writer.addBndBox(bndbox[0], bndbox[1], bndbox[2], bndbox[3], label, difficult)

        writer.save(targetFile=filename)

        return True

    @staticmethod
    def convertPoints2BndBox(points):
        '''
        Copied from the LabelFile class in labelImg.
        :param points: list of points
        :return: tuple containing the bounding box for the list of points with this
                format (int(xmin), int(ymin), int(xmax), int(ymax)).
        '''

        xmin = float('inf')
        ymin = float('inf')
        xmax = float('-inf')
        ymax = float('-inf')
        for p in points:
            x = p[0]
            y = p[1]
            xmin = min(x, xmin)
            ymin = min(y, ymin)
            xmax = max(x, xmax)
            ymax = max(y, ymax)

        # Martin Kersner, 2015/11/12
        # 0-valued coordinates of BB caused an error while
        # training faster-rcnn object detector.
        if xmin < 1:
            xmin = 1

        if ymin < 1:
            ymin = 1

        return (int(xmin), int(ymin), int(xmax), int(ymax))



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from . import rcallback, rcbloadtimeseries, rcbsavetimeseries, rcbloadannotations, rcbimportsiestaannotations, \
    rcbsaveannotations, rcbsaveannotationsXML, rcbzoom, rcbnextframe, rcbprevframe, rcbnoann, rcbfiducials, \
    rcbrectangles, rcbpolylines, rcblivewire, rcbdeletecurrentann, rcbdeleteallann, rcbmeasurepoly, rcbsetbrushsize, \
    rcbcrop, rcbsegment, rcbmovepolyline, rcbsaveimage, rcbtimeslider, rcbhideann, rcbdisplayfiducialids, \
    rcbexportsiestaannotations, rcbimage, rcbexportpolylineannotations, rcbsetcwd, rcbbatchprojectconcatenate, \
    rcbloadclassifier, rcbsaveclassifier, rcbcreateSVM, rcbapplyclassifier, rcbnonmax, rcbabout, \
    rcbdeletefiducialsinoutpoly, rcbtrackfiducials, rcbframespersec, rcbsavedisplay, rcbquit, rcbkymograph, \
    rcbcreateLR, rcbcopypastepolyline

__all__ = [rcallback.RCallback, rcbloadtimeseries.RCBLoadTimeSeries, rcbsavetimeseries.RCBSaveTimeSeries,
           rcbloadannotations.RCBLoadAnnotations,
           rcbimportsiestaannotations.RCBImportSIESTAAnnotations, rcbsaveannotations.RCBSaveAnnotations, rcbzoom.RCBZoom,
           rcbnextframe.RCBNextFrame, rcbprevframe.RCBPrevFrame, rcbnoann.RCBNoAnn, rcbfiducials.RCBFiducials,
           rcbrectangles.RCBRectangles, rcbpolylines.RCBPolylines, rcblivewire.RCBLiveWire,
           rcbdeletecurrentann.RCBDeleteCurrentAnn, rcbdeleteallann.RCBDeleteAllAnn, rcbmeasurepoly.RCBMeasurePoly,
           rcbsaveannotationsXML.RCBSaveAnnotationsXML, rcbsetbrushsize.RCBSetBrushSize, rcbcrop.RCBCrop,
           rcbsegment.RCBSegment, rcbmovepolyline.RCBMovePolyline, rcbsaveimage.RCBSaveImage, rcbtimeslider.RCBTimeSlider,
           rcbhideann.RCBHideAnn, rcbdisplayfiducialids.RCBDisplayFiducialIDs,
           rcbexportsiestaannotations.RCBExportSIESTAAnnotations, rcbimage.RCBImage,
           rcbexportpolylineannotations.RCBExportPolylineAnnotations, rcbsetcwd.RCBSetCWD,
           rcbbatchprojectconcatenate.RCBBatchProjectConcat, rcbloadclassifier.RCBLoadClassifier,
           rcbsaveclassifier.RCBSaveClassifier, rcbcreateSVM.RCBCreateSVM, rcbapplyclassifier.RCBApplyClassifier,
           rcbnonmax.RCBNonMax, rcbabout.RCBAbout, rcbdeletefiducialsinoutpoly.RCBDeleteFiducialsPoly,
           rcbtrackfiducials.RCBTrackFiducials, rcbframespersec.RCBFramesPerSec, rcbsavedisplay.RCBSaveDisplay,
           rcbquit.RCBQuit, rcbkymograph.RCBKymograph, rcbcreateLR.RCBCreateLR, rcbcopypastepolyline.RCBCopyPastePolyline]



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
from .rcallback import RCallback
from PyQt5 import QtWidgets
import os
import scipy.io


class RCBExportSIESTAAnnotations(RCallback):
    def cbExportSIESTAAnnotations(self, filename=''):
        """
        See https://www.mathworks.com/help/matlab/matlab_external/handling-data-returned-from-python.html
        for details.

        Something important here is to send float arrays to Matlab, otherwise there are errors when
        SIESTA tries to conduct certain operations on the arrays.

        :param filename:
        :return:
        """
        fname = ''

        # Get file name.
        if filename == '' or filename is False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            defaultname, _ = os.path.splitext(os.path.basename(self.pjs.filename))

            fname = QtWidgets.QFileDialog.getSaveFileName(None, 'Export SIESTA annotations ...',
                                                          self.pjs.cwd + os.sep + defaultname,
                                                          filter='SIESTA annotations (*.mat)')  # fname[0] is the full filename, fname[1] is the filter used.
            # If cancel ...
            if fname == '':
                return False

            # else ...
            else:
                filename = fname[0]

                # Find max number of fiducials and polygons in any frame.
                rnfiducials = numpy.zeros(self.pjs.n_frames, dtype=numpy.float)
                max_n_polylines = 0
                for iSlice, thefiducials in enumerate(self.pjs.fiducials):
                    rnfiducials[iSlice] = len(thefiducials)
                    if len(self.pjs.polylines[iSlice]) > max_n_polylines:
                        max_n_polylines = len(self.pjs.polylines[iSlice])

                # Create the fiducial array to send to Matlab.
                max_n_fiducials = numpy.amax(rnfiducials)
                rfiducials = -1. * numpy.ones([numpy.int(max_n_fiducials), 3, self.pjs.n_frames], dtype=numpy.float)

                # rpolylines is a list, which in Matlab is a cell.
                # The right dimensions are (max_n_fiducials, 1, self.pjs.n_frames)
                # Need to add the singlet second dimensions. The order of the square brackets is critical here.
                rpolylines = [[[[] for iSlice in range(self.pjs.n_frames)]] for iPol in range(max_n_polylines)]

                for iSlice, thefiducials in enumerate(self.pjs.fiducials):
                    if thefiducials:
                        thefiducials_array = numpy.asarray(thefiducials, dtype=numpy.float)
                        rfiducials[0:thefiducials_array.shape[0], 0:2, iSlice] = thefiducials_array
                        rfiducials[0:thefiducials_array.shape[0], 2, iSlice] = numpy.float(iSlice)

                    thepolylines = self.pjs.polylines[iSlice]

                    if thepolylines:
                        for iPoly, thePoly in enumerate(thepolylines):
                            theintpoly = [[numpy.float(pnt.x()), numpy.float(pnt.y())] for pnt in thePoly]

                            if theintpoly != []:
                                theintpoly.append(theintpoly[0])

                            rpolylines[iPoly][0][iSlice] = theintpoly

                imsize = [numpy.float(self.pjs.width), numpy.float(self.pjs.height), numpy.float(self.pjs.n_frames)]

                # Build dictionary with the variables. Dictionaries are converted to structs
                ud = {}
                ud['rfiducials'] = rfiducials
                ud['rpolygons'] = rpolylines
                ud['rnfiducials'] = rnfiducials
                ud['imsize'] = imsize

                # Open file name and save annotations.
                try:
                    scipy.io.savemat(filename, {'ud': ud}, appendmat=True)
                except (IOError, OSError) as ex:
                    print(ex)
                    return False

        # Modify current path.
        self.pjs.cwd = os.path.dirname(filename)

        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pyjamas.rcallbacks.rcallback import RCallback


# todo: merge the two methods into a single one that can accept one parameter.
class RCBTimeSlider(RCallback):
    def cbTimeSlider(self) -> bool:
        self.pjs.curslice = self.pjs.timeSlider.value()-1
        self.pjs.imagedata = self.pjs.slices[self.pjs.curslice]

        self.pjs.displayData()

        return True

    def cbGoTo(self, slice_index: int) -> bool:
        """

        :param slice_index: negative values start start pointing from the last slice (-1 being the last one).
        :return:
        """
        if slice_index >= self.pjs.n_frames:
            return False

        if slice_index < 0:
            slice_index = self.pjs.n_frames+slice_index
            if slice_index < 0:
                return False

        self.pjs.curslice = slice_index
        self.pjs.imagedata = self.pjs.slices[self.pjs.curslice]
        self.pjs.timeSlider.setValue(slice_index+1)

        self.pjs.displayData()

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt5 import QtCore, QtWidgets

import pyjamas.pjscore as pjscore
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rutils import RUtils


class RCBCopyPastePolyline(RCallback):

    PIX_SHIFT: int = 2  # x, y shift to use when pasting polylines with a shift with respect to original position.

    def cbCopyPolyline(self) -> bool:
        self.pjs.annotation_mode = pjscore.PyJAMAS.copy_polyline
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.CrossCursor)

        return True

    def cbPastePolyline(self, paste_shifted: bool = False) -> object:
        if self.pjs._copied_poly_ is None or self.pjs._copied_poly_ == []:
            return False

        if paste_shifted:
            copied_poly = (RUtils.qpolygonf2ndarray(self.pjs._copied_poly_) + RCBCopyPastePolyline.PIX_SHIFT).tolist()
        else:
            copied_poly = RUtils.qpolygonf2list(self.pjs._copied_poly_)

        self.pjs.addPolyline(copied_poly, self.pjs.curslice)

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .rcallback import RCallback
from PyQt5 import QtWidgets


class RCBFramesPerSec(RCallback):
    def cbFramesPerSec(self, fps: int=None) -> int:
        '''
        Sets the size of the brush used to paint polygons.
        :return: int (selected brush size or -1 if cancelled)
        '''

        thefps: int = 0
        ok_flag: bool = None

        if fps is not None and fps is not False:
            thefps = fps
            ok_flag = True
        else:
            # Read user input for fps.
            thefps, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set frames per second: ', 'Enter frames per second: ',
                                                            self.pjs.fps, 1)

        
        if ok_flag and thefps > 0:
            self.pjs.fps = thefps

            return True

        else:
            return False



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

import numpy
from PyQt5 import QtWidgets

from .rcallback import RCallback
from pyjamas.rimage.rimcore import rimage
from pyjamas.rimage.rimutils import rimutils


class RCBLoadTimeSeries(RCallback):
    def cbLoadTimeSeries(self, filename: str = '') -> bool:  # function cancel = cbLoadTimeSeries(fig, filename, in)
        """
        Loads a grayscale, multi-page TIFF.
        :param filename: file to open.
        :return: boolean (True if the image was loaded with no problems, False otherwise)
        """

        # Get file name.
        if filename == '' or filename is False:
            fname = QtWidgets.QFileDialog.getOpenFileName(None, 'Load grayscale image ...', self.pjs.cwd,
                                                          filter='All files (*)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        _, ext = os.path.splitext(filename)

        if ext not in rimage.image_extensions:
            self.pjs.statusbar.showMessage("That does not look like an image.")
            return False

        # Read image.
        self.pjs.slices = rimutils.read_stack(filename)

        # MISSING: error checking.

        # Store current working directory.
        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.
        self.pjs.filename = filename

        # Initialize image and display.
        self.pjs.initImage()

        return True

    def cbLoadArray(self, image: numpy.ndarray) -> bool:
        if image is False or image is None:
            return False

        self.pjs.slices = image

        self.pjs.filename = 'animage'

        self.pjs.initImage()




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Tuple

import matplotlib
matplotlib.use('Qt5Agg')  # https://stackoverflow.com/questions/21784641/installation-issue-with-matplotlib-python
import matplotlib.pyplot as plt
import numpy
from scipy.sparse.csgraph import dijkstra
from scipy.misc import toimage

from .rimutils import rimutils


# SHOULD THIS BE CALLED RImageOps and all methods be static (similar to PIL.ImageOps)
class rimage:
    """
    A class to do image operations.

    The image is stored as a numpy array (image_array).

    Because numpy arrays are accessed by row and col, all methods implemented here will use the row, col convention even to
    represent points.
    """

    image_extensions: Tuple[str] = ('.tif', '.tiff', '.jpg')

    def __init__(self, image_in):
        self.image_array:numpy.array = image_in

    # Return a tuple with all the points between thesrc and thedst. thesrc and thedst are in (col=x, row=y) format
    def livewire(self, thesrc, thedst, margin):
        # Copy parameters. This is critical. I had a headache, because the function was modifying the parameter values.
        src = thesrc.copy()
        dst = thedst.copy()

        # Crop image. Note that image_array.shape returns size in (rows, columns) format. But src and dst are in (col=x, row=y) format.
        minx = max(0, min(src[0], dst[0]) - margin)
        maxx = min(self.image_array.shape[1] - 1, max(src[0], dst[0]) + margin) + 1
        miny = max(0, min(src[1], dst[1]) - margin)
        maxy = min(self.image_array.shape[0] - 1, max(src[1], dst[1]) + margin) + 1

        # Crop image.
        if self.image_array.ndim == 3:
            im = self.image_array[miny:maxy, minx:maxx, 0].copy()
        else:
            im = self.image_array[miny:maxy, minx:maxx].copy()

        # Adjust source and destination coordinates.
        src[0] -= minx
        src[1] -= miny
        dst[0] -= minx
        dst[1] -= miny

        src_ind, dst_ind = rimutils.sub2ind(im.shape, numpy.array([src[1], dst[1]]), numpy.array([src[0], dst[0]]))

        # Make sure the cropped image is a matrix and invert it so that high intensities become low ones, and we can
        # search for the minimal cost path.
        im[:] = numpy.max(im) - im[:]  # This modifies the array in place, which avoids having to reallocate the memory
        # (which is slow).
        # Doing im = numpy.max(im) - im produces the same result, but reallocates the memory.

        # Create graph.
        matrix_sparse, _ = rimutils.makeGraph(im)

        distances, predecessors = dijkstra(matrix_sparse, directed=True, indices=src_ind, return_predecessors=True)

        path = []

        i = dst_ind
        while i != src_ind:
            path.append(i)
            i = predecessors[i]

        path.append(src_ind)
        path = path[::-1]

        path = rimutils.ind2sub(im.shape, numpy.array(path))

        # Generate coordinates.
        return [[x+minx, y+miny] for x, y in zip(path[:][1], path[:][0])]

    def stretch(self, low: int = 0, high: int = 100, minimum: int = 0, maximum: int = 255) -> numpy.ndarray:
        """
        Linear stretch of the contrast of an image.
        :param low: lower percentile of the image to be mapped to the minimum value (0)
        :param high: higher percentile (100)
        :param minimum: minimum pixel value in the resulting image (0)
        :param maximum: maximum pixel value (255)
        :return: image_out, an image array.

        Has a sister method in rimutils.
        """

        image_out: numpy.array = []

        if high == low:
            return self.image_array

        if low > high:
            high, low = low, high

        # Define output variable.
        image_out = self.image_array

        # Find the appropriate pixel values for the low and high thresholds.
        if low == 0:
            low = numpy.min(self.image_array)
        else:
            low = numpy.percentile(self.image_array, low)

        if high == 100:
            high = numpy.max(self.image_array)
        else:
            high = numpy.percentile(self.image_array, high)

        # Determine the scaling factor.
        sc = (maximum - minimum) / (high - low)

        # Linear stretch of image_in.
        if low != 0:
            image_out = image_out - low

        if sc != 1:
            image_out = image_out * sc

        if minimum != 0:
            image_out = image_out + minimum

        return image_out

    def show(self):
        plt.imshow(toimage(self.image_array))
        plt.show()

        return


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
from PyQt5 import QtWidgets

import pyjamas.dialogs as dialogs
from .rcallback import RCallback
from pyjamas.rimage.rimclassifier.rimclassifier import rimclassifier


class RCBNonMax(RCallback):
    def cbNonMaxSuppression(self, parameters: dict = None, first_tp: int = None, last_tp: int = None) -> bool:  # Handle IO errors.
        continue_flag = True

        if parameters is None or parameters is False:
            dialog = QtWidgets.QDialog()
            ui = dialogs.nonmax_suppr.NonMaxDialog(self.pjs)
            ui.setupUi(dialog)
            dialog.exec_()
            dialog.show()

            continue_flag = dialog.result() == QtWidgets.QDialog.Accepted

            if continue_flag:
                parameters = ui.parameters()

            dialog.close()

        if continue_flag:
            if (first_tp is None or first_tp is False) and (last_tp is None or last_tp is False):
                dialog = QtWidgets.QDialog()
                ui = dialogs.timepoints.TimePointsDialog()
                ui.setupUi(dialog, dialogs.timepoints.TimePointsDialog.firstSlice,
                           dialogs.timepoints.TimePointsDialog.lastSlice)

                dialog.exec_()
                dialog.show()

                continue_flag = dialog.result() == QtWidgets.QDialog.Accepted

                if continue_flag:
                    first_tp, last_tp = ui.parameters()

                dialog.close()

            if first_tp <= last_tp:
                theslicenumbers = numpy.arange(first_tp - 1, last_tp, dtype=int)
            else:
                theslicenumbers = numpy.arange(last_tp - 1, first_tp, dtype=int)

            self.pjs.batch_classifier.non_max_suppression(
                parameters.get('prob_threshold', rimclassifier.DEFAULT_PROB_THRESHOLD),
                parameters.get('iou_threshold', rimclassifier.DEFAULT_IOU_THRESHOLD),
                parameters.get('max_num_objects', rimclassifier.DEFAULT_MAX_NUM_OBJECTS),
                theslicenumbers
            )

            for index in theslicenumbers:
                self.pjs._cbDeleteCurrentAnn.cbDeleteCurrentAnn(index)
                self.pjs._cbApplyClassifier.add_classifier_boxes(self.pjs.batch_classifier.box_arrays[index][self.pjs.batch_classifier.good_box_indices[index]], index)

            self.pjs.repaint()

            return True

        else:
            return False




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
import os
from PyQt5 import QtWidgets
import scipy.io

from pyjamas.rcallbacks.rcallback import RCallback


class RCBImportSIESTAAnnotations(RCallback):
    def cbImportSIESTAAnnotations(self, filename='', image_file: str = None):
        # Get file name.
        if filename == '' or filename is False:  # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getOpenFileName(None, 'Import SIESTA annotations ...', self.pjs.cwd,
                                                          filter='SIESTA annotations (*.mat)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        # Open file name and read annotations.
        try:
            matlabVars = scipy.io.loadmat(filename, struct_as_record=False)
        except (IOError, OSError) as ex:
            print(ex)
            return False

        ud = matlabVars['ud'][0][0]

        fiducials = numpy.transpose(ud.rfiducials, [2, 0, 1])
        polyline_list = numpy.transpose(ud.rpolygons, [2, 0, 1])

        if image_file == '':  # Create a blank image.
            # Find maxx, maxy, and maxz, and create a numpy.ndarray with those dimensions + 1.
            maxz = len(fiducials)
            maxx = 0
            maxy = 0

            # Fiducials and polylines are stored as (x, y) coordinates.
            # numpy.ndarrays are created with (slices, rows, cols).
            for slice_fiducials in fiducials:
                for afiducial in slice_fiducials:
                    maxx = max(maxx, afiducial[0])
                    maxy = max(maxy, afiducial[1])

            for slice_polylines in polyline_list:
                for apolyline in slice_polylines:
                    # For some reason, loadmat puts the polygons within a list.
                    if len(apolyline) == 1:
                        for afiducial in apolyline[0]:
                            if len(afiducial) > 0:
                                maxx = max(maxx, afiducial[0])
                                maxy = max(maxy, afiducial[1])

            virtual_image = numpy.zeros((maxz, int(maxy + 1), int(maxx + 1)), dtype=int)
            self.pjs._cbLoadTS.cbLoadArray(virtual_image)

        elif image_file is not False and image_file is not None:
            self.pjs._cbLoadTS.cbLoadTimeSeries(image_file)

        self.pjs.fiducials = [[] for i in range(self.pjs.n_frames)]

        for slice_fiducials in fiducials:
            for afiducial in slice_fiducials:
                if afiducial[-1] >= 0:
                    self.pjs.addFiducial(int(afiducial[0]), int(afiducial[1]), int(afiducial[2]))

        self.pjs.polylines = [[] for i in range(self.pjs.n_frames)]

        for z, slice_polylines in enumerate(polyline_list):
            for apolyline in slice_polylines:
                # For some reason, loadmat puts the polygons within a list.
                if len(apolyline[0]) > 1:
                    self.pjs.addPolyline(apolyline[0].tolist(), z)

        self.pjs.repaint()

        # Modify current path.
        self.pjs.cwd = filename[0:filename.rfind(os.sep)]

        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gzip
import os
import pickle

from PyQt5 import QtWidgets
from sklearn.linear_model.logistic import LogisticRegression
from sklearn.svm.classes import SVC

from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimclassifier.rimclassifier import rimclassifier
from pyjamas.rimage.rimclassifier.rimlr import lr
from pyjamas.rimage.rimclassifier.rimsvm import svm



class RCBLoadClassifier(RCallback):
    def cbLoadClassifier(self, filename: str = '') -> bool:  # Handle IO errors
        # Get file name.
        if filename == '' or filename is False: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            fname = QtWidgets.QFileDialog.getOpenFileName(None, 'Load classifier ...', self.pjs.cwd,
                                                          filter='PyJAMAS classifier (*' + rimclassifier.CLASSIFIER_EXTENSION + ')')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        # Open file name and read classifier.
        fh = None
        theparameters: dict = None

        try:
            fh = gzip.open(filename, "rb")
            theparameters = pickle.load(fh)

        except (IOError, OSError) as ex:
            if fh is not None:
                fh.close()

            print(ex)
            return False


        if type(theparameters['classifier']) is SVC:
            self.pjs.batch_classifier.image_classifier = svm(theparameters)
        elif type(theparameters['classifier']) is LogisticRegression:
            self.pjs.batch_classifier.image_classifier = lr(theparameters)
        else:
            self.pjs.statusbar.showMessage(f"Wrong classifier type.")
            return False

        # Modify current path.
        self.pjs.cwd = filename[0:filename.rfind(os.sep)]

        self.pjs.statusbar.showMessage(f"Classifier {filename} loaded.")

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

from PyQt5 import QtWidgets

from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimutils import rimutils


class RCBSaveImage(RCallback):
    def cbSaveImage(self, filename: str=False, x_range: tuple=False, y_range: tuple=False, z_range: tuple=False) -> bool:  # function cancel = cbLoadTimeSeries(fig, filename, in)
        """

        :param filename: '' for automated naming based on coordinates. If not provided, a dialog will open.
        :param x_range: tuple containing the min and max X values to save. If False, take the coordinates of the first polygon.
        :param y_range: tuple containing the min and max Y values to save. If False, take the coordinates of the first polygon.
        :param z_range: tuple containing the min and max Z values to save. If False, use just the current Z.
        :return: True if the image was properly saved; False if the save cannot be completed.
        """

        # Check X, Y, and Z parameters or assign default values.
        if z_range is False:
            z_range = tuple([self.pjs.curslice, self.pjs.curslice+1])

        if x_range is False:
            thepolylines = self.pjs.polylines[self.pjs.curslice]
            if thepolylines != [] and thepolylines[0] != []:
                thepolyline = self.pjs.polylines[self.pjs.curslice][0].boundingRect()
                x_range = tuple([int(thepolyline.x()), int(thepolyline.x()+thepolyline.width()+1)])
            else:
                return False

        if y_range is False and thepolylines != [] and thepolylines[0] != []:
            thepolylines = self.pjs.polylines[self.pjs.curslice]
            if thepolylines != [] and thepolylines[0] != []:
                thepolyline = self.pjs.polylines[self.pjs.curslice][0].boundingRect()
                y_range = tuple([int(thepolyline.y()), int(thepolyline.y() + thepolyline.height() + 1)])
            else:
                return False

        # Get file name.
        if filename is False:
            fname: tuple = QtWidgets.QFileDialog.getSaveFileName(None, 'Save time series ...', self.pjs.cwd,
                                                          filter='TIFF files (*.tif *.tiff)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            filename = self.pjs.filename + '_X' + str(x_range[0]) + '_' + str(x_range[1]-1) + '_Y' + str(y_range[0]) + \
                       '_' + str(y_range[1]-1) + '_Z' + str(z_range[0]) + '_' + str(z_range[1]-1) + '.tif'

        # Save image.
        rimutils.write_stack(filename,
                             self.pjs.slices[z_range[0]:z_range[1], y_range[0]:y_range[1], x_range[0]:x_range[1]])
        self.pjs.statusbar.showMessage('Image saved: ' + filename)

        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.

        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy

import pyjamas.pjscore as pjscore
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimutils import rimutils


class RCBKymograph(RCallback):
    """
    Creates the kymograph of an image.
    """

    def cbKymograph(self, coordinates: numpy.ndarray = None, new_window: bool = True) -> bool:
        """

        :param coordinates: ndarray with two columns containing the x, y coordinates of the region to crop.
        :param new_window: True if the cropped image should open in a new PyJAMAS session, False if the current window
        should be used instead.
        :return: False if no coordinates are provided, True otherwise.
        """
        if (not coordinates) | (coordinates is None) | (coordinates == []):
            thepolylines = self.pjs.polylines[self.pjs.curslice]

            if thepolylines != [] and thepolylines[0] != []:
                thepolyline = thepolylines[0].boundingRect()
                minx, miny, maxx, maxy = thepolyline.getCoords()

            else:
                return False

        else:
            # This is here mainly to crop around non-rectangular polylines.
            minx, miny = numpy.min(coordinates, axis=0)
            maxx, maxy = numpy.max(coordinates, axis=0)

        # Make sure you are working with integers to prevent errors when slicing the original image.
        coordinates: numpy.ndarray = numpy.asarray([[minx, miny], [maxx, maxy]])

        # Make kymograph.
        thekymo = rimutils.kymograph(self.pjs.slices, coordinates)

        # Open in current window or new window?
        if not new_window:
            self.pjs._cbLoadTS.cbLoadArray(thekymo)
        else:
            # Create a new window containing the cropped image.
            new_pjs: pjscore.PyJAMAS = pjscore.PyJAMAS()
            new_pjs._cbLoadTS.cbLoadArray(thekymo)

        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pyjamas.pjscore as pjscore
from pyjamas.rcallbacks.rcallback import RCallback


class RCBZoom(RCallback):
    def cbZoom(self):
        self.pjs.gView.scale(1./pjscore.PyJAMAS.zoom_factors[self.pjs.zoom_index],
                             1./pjscore.PyJAMAS.zoom_factors[self.pjs.zoom_index])
        self.pjs.zoom_index = (self.pjs.zoom_index+1) % len(pjscore.PyJAMAS.zoom_factors)
        #print(str(self.pjs.zoom_index))
        self.pjs.gView.scale(pjscore.PyJAMAS.zoom_factors[self.pjs.zoom_index], pjscore.PyJAMAS.zoom_factors[self.pjs.zoom_index])
        self.pjs.MainWindow.resize(self.pjs.width * self.pjs.zoom_factors[self.pjs.zoom_index],
                                   self.pjs.height * self.pjs.zoom_factors[self.pjs.zoom_index] + 60)
        self.pjs.statusbar.showMessage(str(self.pjs.curslice + 1) + '/' + str(self.pjs.n_frames) + ' zoom: ' + str(pjscore.PyJAMAS.zoom_factors[self.pjs.zoom_index]) + 'x')

        return True





"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

from PyQt5 import QtWidgets, QtGui
from PyQt5.QtCore import QRect, QPointF

from pyjamas.rcallbacks.rcallback import RCallback


class RCBSaveDisplay(RCallback):
    def cbSaveDisplay(self, filename: str = None) -> bool:
        """

        :param filename: '' for automated naming based on coordinates. If not provided, a dialog will open.
        :return: True if the display was properly saved; False if the save cannot be completed.
        """

        # Get file name.
        if filename is False or filename is None:
            fname: tuple = QtWidgets.QFileDialog.getSaveFileName(None, 'Save time series ...', self.pjs.cwd,
                                                          filter='TIFF files (*.tif *.tiff)')  # fname[0] is the full filename, fname[1] is the filter used.
            filename = fname[0]
            if filename == '':
                return False

        elif filename == '':
            filename = self.pjs.filename

        # Save image.
        self.save_display(filename)

        self.pjs.statusbar.showMessage('Display saved: ' + filename)

        self.pjs.cwd = os.path.dirname(filename)  # Path of loaded image.

        return True

    def save_display(self, filename: str = None) -> QtGui.QPixmap:

        #pix_map: QtGui.QPixmap = self.pjs.gView.grab(self.pjs.gView.sceneRect().toRect())
        #pix_map: QtGui.QPixmap = self.pjs.gView.grab(QRect(0, 0, int(self.pjs.width * self.pjs.zoom_factors[self.pjs.zoom_index]), int(self.pjs.height * self.pjs.zoom_factors[self.pjs.zoom_index])))
        pix_map: QtGui.QPixmap = self.pjs.gView.grab()
        # todo: For some reason, the pix_map returned by grab is twice as large as it should (??).
        pix_map = pix_map.scaled(pix_map.width()/2, pix_map.height()/2)

        if filename and filename is not None and filename != '':
            pix_map.save(filename)

        return pix_map

    def export_movie(self, filename: str = None) -> bool:

        pass
